import argparse
import imageio.v3 as iio
import numpy as np

from pathlib import Path
from tqdm.auto import tqdm

from piscis import Piscis
from piscis.training import train_model


def main():

    """Main function for the Piscis command-line interface."""

    parser = argparse.ArgumentParser(description="Piscis Algorithm")
    subparsers = parser.add_subparsers(dest='command')
    formatter_class = argparse.ArgumentDefaultsHelpFormatter

    # Predict subparser.
    predict_parser = subparsers.add_parser('predict', formatter_class=formatter_class,
                                           help="Predict spots using a trained SpotsModel.")
    predict_parser.add_argument('input_path', type=str,
                                help="Path to image or stack of images.")
    predict_parser.add_argument('output_path', type=str,
                                help="Path to save predicted spots as CSV.")
    predict_parser.add_argument('--model-name', type=str, default='20230905',
                                help="Model name.")
    predict_parser.add_argument('--batch-size', type=int, default=4,
                                help="Batch size for the CNN.")
    predict_parser.add_argument('--stack', action='store_true',
                                help="Whether the input is a stack of images.")
    predict_parser.add_argument('--scale', type=float, default=1.0,
                                help="Scale factor for rescaling the input.")
    predict_parser.add_argument('--threshold', type=float, default=1.0,
                                help="Spot detection threshold. Can be interpreted as the minimum number of fully "
                                     "confident pixels necessary to identify a spot.")
    predict_parser.add_argument('--min-distance', type=int, default=1,
                                help="Minimum distance between spots.")

    # Train subparser.
    train_parser = subparsers.add_parser('train', formatter_class=formatter_class,
                                         help="Train a SpotsModel")
    train_parser.add_argument('model_name', type=str,
                              help="Name of a new or existing model.")
    train_parser.add_argument('dataset_path', type=str,
                              help="Path to the directory containing training and validation datasets.")
    train_parser.add_argument('--initial-model-name', type=str, default=None,
                              help="Name of an existing model to initialize the weights.")
    train_parser.add_argument('--adjustment', type=str, default='standardize',
                              help="Adjustment type applied to images. Supported types are 'normalize' and "
                                   "'standardize'.")
    train_parser.add_argument('--input-size', type=int, nargs=2, default=(256, 256),
                              help="Size of the input images used for training.")
    train_parser.add_argument('--random-seed', type=int, default=0,
                              help="Random seed used for initialization and training.")
    train_parser.add_argument('--batch-size', type=int, default=4,
                              help="Batch size for training.")
    train_parser.add_argument('--learning-rate', type=float, default=0.2,
                              help="Learning rate for the optimizer.")
    train_parser.add_argument('--weight-decay', type=float, default=1e-4,
                              help="Strength of the weight decay regularization.")
    train_parser.add_argument('--dropout-rate', type=float, default=0.2,
                              help="Dropout rate at skip connections.")
    train_parser.add_argument('--epochs', type=int, default=400,
                              help="Number of epochs to train the model for.")
    train_parser.add_argument('--warmup-fraction', type=float, default=0.05,
                              help="Fraction of epochs for learning rate warmup.")
    train_parser.add_argument('--decay-fraction', type=float, default=0.5,
                              help="Fraction of epochs for learning rate decay.")
    train_parser.add_argument('--decay-transitions', type=int, default=10,
                              help="Number of times to decay the learning rate.")
    train_parser.add_argument('--decay-factor', type=float, default=0.5,
                              help="Multiplicative factor of each learning rate decay transition.")
    train_parser.add_argument('--dilation-iterations', type=int, default=1,
                              help="Number of iterations to dilate ground truth labels to minimize class imbalance and "
                                   "misclassifications due to minor offsets.")
    train_parser.add_argument('--max-distance', type=float, default=3.0,
                              help="Maximum distance for matching predicted and ground truth displacement vectors.")
    train_parser.add_argument('--l2-loss-weight', type=float, default=0.25,
                              help="Weight for the L2 loss term.")
    train_parser.add_argument('--bce-loss-weight', type=float, default=0.0,
                              help="Weight for the bce loss term.")
    train_parser.add_argument('--dice-loss-weight', type=float, default=0.0,
                              help="Weight for the dice loss term.")
    train_parser.add_argument('--focal-loss-weight', type=float, default=0.0,
                              help="Weight for the focal loss term.")
    train_parser.add_argument('--smoothf1-loss-weight', type=float, default=1.0,
                              help="Weight for the smoothf1 loss term.")
    train_parser.add_argument('--save-checkpoints', default=True, action=argparse.BooleanOptionalAction,
                              help="Save checkpoints during training.")

    args = parser.parse_args()

    if args.command == 'train':

        # Create the loss weights dictionary.
        loss_weights = {}
        if args.l2_loss_weight > 0:
            loss_weights['l2'] = args.l2_loss_weight
        if args.bce_loss_weight > 0:
            loss_weights['bce'] = args.bce_loss_weight
        if args.dice_loss_weight > 0:
            loss_weights['dice'] = args.dice_loss_weight
        if args.focal_loss_weight > 0:
            loss_weights['focal'] = args.focal_loss_weight
        if args.smoothf1_loss_weight > 0:
            loss_weights['smoothf1'] = args.smoothf1_loss_weight

        # Train the model.
        train_model(
            model_name=args.model_name,
            dataset_path=args.dataset_path,
            initial_model_name=args.initial_model_name,
            adjustment=args.adjustment,
            input_size=args.input_size,
            random_seed=args.random_seed,
            batch_size=args.batch_size,
            learning_rate=args.learning_rate,
            weight_decay=args.weight_decay,
            dropout_rate=args.dropout_rate,
            epochs=args.epochs,
            warmup_fraction=args.warmup_fraction,
            decay_fraction=args.decay_fraction,
            decay_transitions=args.decay_transitions,
            decay_factor=args.decay_factor,
            dilation_iterations=args.dilation_iterations,
            max_distance=args.max_distance,
            loss_weights=loss_weights,
            save_checkpoints=args.save_checkpoints
        )

    elif args.command == 'predict':

        input_path = Path(args.input_path)
        output_path = Path(args.output_path)
        output_path.mkdir(parents=True, exist_ok=True)

        # Get a list of image paths.
        image_paths = []
        if input_path.is_dir():
            for path in input_path.glob("*"):
                if path.is_file():
                    image_paths.append(path)
        else:
            image_paths.append(input_path)

        # Create the Piscis model.
        model = Piscis(model_name=args.model_name, batch_size=args.batch_size, input_size=None)

        # Loop over image paths.
        for image_path in tqdm(image_paths):

            # Load an image and predict spots.
            image = iio.imread(image_path)
            coords = model.predict(
                image,
                stack=args.stack,
                scale=args.scale,
                threshold=args.threshold,
                min_distance=args.min_distance,
                intermediates=False
            )

            # Save the predicted spots to a CSV file.
            np.savetxt(output_path / f'{image_path.stem}.csv', coords, delimiter=',')

    else:

        parser.print_help()


if __name__ == '__main__':
    main()



import jax
import numpy as np

from flax import linen as nn
from typing import Tuple, Union

from piscis.networks.efficientnetv2 import build_efficientnetv2
from piscis.networks.fpn import FPN

BLOCK_ARGS = [
    {
        "kernel_size": (3, 3),
        "num_repeat": 4,
        "features_in": 32,
        "features_out": 32,
        "expand_ratio": 1,
        "se_ratio": 0.0,
        "strides": 1,
        "conv_type": 1,
        "pool": None
    }, {
        "kernel_size": (3, 3),
        "num_repeat": 4,
        "features_in": 32,
        "features_out": 64,
        "expand_ratio": 2,
        "se_ratio": 0.0,
        "strides": 1,
        "conv_type": 1,
        "pool": "max"
    }, {
        "conv_type": 0,
        "expand_ratio": 4,
        "features_in": 64,
        "kernel_size": (3, 3),
        "num_repeat": 4,
        "features_out": 128,
        "se_ratio": 0.25,
        "strides": 1,
        "pool": "max"
    }, {
        "conv_type": 0,
        "expand_ratio": 4,
        "features_in": 128,
        "kernel_size": (3, 3),
        "num_repeat": 4,
        "features_out": 256,
        "se_ratio": 0.25,
        "strides": 1,
        "pool": "max"
    }
]


class SpotsModel(nn.Module):

    """Spot detection model.

    Attributes
    ----------
    style : bool
        Whether to use style transfer.
    aggregate : str
        Aggregation mode for the feature pyramid network. Supported modes are 'sum' and 'concatenate'.
    dropout_rate : float
        Dropout rate at skip connections.
    """

    style: bool = True
    aggregate: str = 'sum'
    dropout_rate: float = 0.2

    @nn.compact
    def __call__(
            self,
            x: jax.Array,
            train: bool = True,
            return_style: bool = False
    ) -> Union[Tuple[jax.Array, jax.Array], Tuple[jax.Array, jax.Array, jax.Array]]:

        encoder = build_efficientnetv2(
            blocks_args=BLOCK_ARGS,
            model_name='EfficientNetV2XS',
            width_coefficient=1.0,
            depth_coefficient=1.0,
            stem_strides=1,
            dropout_rate=self.dropout_rate
        )

        x, style = FPN(
            encoder=encoder,
            encoder_levels={0, 1, 2, 3},
            features=3,
            style=self.style,
            aggregate=self.aggregate,
            dropout_rate=self.dropout_rate
        )(x, train=train)
        deltas = x[:, :, :, :2]
        labels = nn.sigmoid(x[:, :, :, 2:3])

        if return_style:
            return deltas, labels, style
        else:
            return deltas, labels


def round_input_size(input_size: Tuple[int, int]) -> Tuple[int, int]:

    """Round SpotsModel input size.

    Parameters
    ----------
    input_size : Tuple[int, int]
        Input size.

    Returns
    -------
    rounded_input_size : Tuple[int, int]
        Rounded input size.
    """

    stride_scale = np.prod([block['strides'] for block in BLOCK_ARGS])
    pool_scale = 2 ** sum((0 if block['pool'] is None else 1 for block in BLOCK_ARGS if block['pool']))
    scale = stride_scale * pool_scale
    rounded_input_size = scale * np.rint(np.array(input_size) / scale).astype(int)
    rounded_input_size = (int(rounded_input_size[0]), int(rounded_input_size[1]))

    return rounded_input_size



from huggingface_hub import HfFileSystem
from pathlib import Path
from typing import Sequence

from piscis.paths import HF_DATASETS_DIR, HF_MODELS_DIR, MODELS_DIR


def download_pretrained_model(
        model_name: str
) -> None:

    """Download a pretrained model from Hugging Face.

    Parameters
    ----------
    model_name : str
        Model name.

    Raises
    ------
    ValueError
        If `model_name` is not found.
    """

    # Create the models directory if necessary.
    MODELS_DIR.mkdir(parents=True, exist_ok=True)

    # Download the model if available.
    fs = HfFileSystem()
    hf_model_path = f'{HF_MODELS_DIR}{model_name}'
    if fs.exists(hf_model_path):
        fs.download(hf_model_path, str(MODELS_DIR / model_name))
    else:
        raise ValueError(f"Model {model_name} is not found.")


def list_pretrained_models() -> Sequence[str]:

    """List pretrained models from Hugging Face.

    Returns
    -------
    pretrained_models : Sequence[str]
        List of pretrained model names.
    """

    fs = HfFileSystem()
    pretrained_models = [path.removeprefix(HF_MODELS_DIR) for path in fs.ls(HF_MODELS_DIR, detail=False)]

    return pretrained_models


def download_dataset(
        dataset_name: str,
        path: str,
        minimal_download: bool = True
) -> None:

    """Download a dataset from Hugging Face.

    Parameters
    ----------
    dataset_name : str
        Dataset name.
    path : str
        Path to save the dataset.
    minimal_download : bool, optional
        Whether to download only the necessary files for training and inference. Default is True.

    Raises
    ------
    ValueError
        If `dataset_name` is not found.
    """

    fs = HfFileSystem()
    hf_dataset_path = f'{HF_DATASETS_DIR}{dataset_name}'
    if fs.exists(hf_dataset_path) and ('/' not in dataset_name):
        path = str(path)
        if minimal_download:
            dataset_path = Path(path) / dataset_name
            dataset_path.mkdir(parents=True, exist_ok=True)
            dataset_path = str(dataset_path)
            for file_path in fs.glob(f'{hf_dataset_path}/*.npz'):
                fs.download(file_path, dataset_path)
        else:
            fs.download(hf_dataset_path, path, recursive=True)
    else:
        raise ValueError(f"Dataset {dataset_name} is not found.")


def list_datasets() -> Sequence[str]:

    """List datasets from Hugging Face.

    Returns
    -------
    datasets : Sequence[str]
        List of dataset names.
    """

    fs = HfFileSystem()
    datasets = [path.removeprefix(HF_DATASETS_DIR) for path in fs.ls(HF_DATASETS_DIR, detail=False) if fs.isdir(path)]

    return datasets



import jax
import jax.numpy as jnp
import numpy as np

from jax import vmap
from jax.lax import dynamic_slice, scan
from numba import njit
from scipy import optimize
from skimage import feature, measure
from typing import Dict, List, Sequence, Tuple


def compute_spot_coordinates(
        deltas: np.ndarray,
        pooled_labels: np.ndarray,
        threshold: float,
        min_distance: int,
) -> np.ndarray:

    """Compute spot coordinates from deltas and labels.

    Parameters
    ----------
    deltas : np.ndarray
        Displacement vectors.
    pooled_labels : np.ndarray
        Pooled labels.
    threshold : float
        Detection threshold between 0 and 1.
    min_distance : int
        Minimum distance between spots.

    Returns
    -------
    coords : np.ndarray
        Coordinates of detected spots.
    """

    # Check if the pooled labels are in a stack.
    stack = pooled_labels.ndim == 3

    if stack:

        # Use connected components to detect spots if pooled labels are in a stack.
        labels = measure.label(pooled_labels > threshold)
        peaks = np.array([region['centroid'] for region in measure.regionprops(labels)], dtype=int)

    else:

        # Use peak local maxima to detect spots if pooled labels are not in a stack.
        peaks = feature.peak_local_max(pooled_labels,
                                       min_distance=min_distance, threshold_abs=threshold, exclude_border=False)

    # Apply deltas to detected spots.
    if len(peaks) > 0:
        if stack:
            coords = peaks + np.pad(deltas[peaks[:, 0], peaks[:, 1], peaks[:, 2]], ((0, 0), (1, 0)))
        else:
            coords = peaks + deltas[peaks[:, 0], peaks[:, 1]]
    else:
        coords = np.empty((0, pooled_labels.ndim), dtype=np.float32)

    return coords


def scanned_sum_pool(
        deltas: jax.Array,
        labels: jax.Array,
        n_iter: int,
        kernel_size: Sequence[int] = (3, 3)
) -> jax.Array:

    """Scanned version of `sum_pool`.

    Parameters
    ----------
    deltas : jax.Array
        Displacement vectors.
    labels : jax.Array
        Binary labels.
    n_iter : int
        Number of iterations.
    kernel_size : Sequence[int], optional
        Kernel size or window size of the sum pooling operation. Default is (3, 3).

    Returns
    -------
    pooled_labels : jax.Array
        Pooled labels after sum pooling for `n_iter` iterations.
    """

    pooled_labels = scan(lambda c, x: _scanned_sum_pool(deltas, c, kernel_size), labels, jnp.empty(n_iter))[0]

    return pooled_labels


vmap_scanned_sum_pool = vmap(scanned_sum_pool, in_axes=(0, 0, None, None))


def _scanned_sum_pool(
        deltas: jax.Array,
        pooled_labels: jax.Array,
        kernel_size: Sequence[int]
) -> Tuple[jax.Array, None]:

    """Single iteration of `scanned_apply_deltas`.

    Parameters
    ----------
    deltas : jax.Array
        Displacement vectors.
    pooled_labels : jax.Array
        Pooled labels carried over from the previous iteration.
    kernel_size : Sequence[int]
        Kernel size or window size of the sum pooling operation.

    Returns
    -------
    pooled_labels : jax.Array
        Pooled labels to carry to the next iteration.
    None
    """

    pooled_labels = sum_pool(deltas, pooled_labels, kernel_size)

    return pooled_labels, None


def sum_pool(
        deltas: jax.Array,
        labels: jax.Array,
        kernel_size: Sequence[int] = (3, 3)
) -> jax.Array:

    """Sum pool labels using deltas.

    Parameters
    ----------
    deltas : jax.Array
        Displacement vectors.
    labels : jax.Array
        Binary labels.
    kernel_size : Sequence[int], optional
        Kernel size or window size of the sum pooling operation. Default is (3, 3).

    Returns
    -------
    pooled_labels : jax.Array
        Pooled labels.
    """

    # Generate an index map.
    i, j = jnp.arange(deltas.shape[0]), jnp.arange(deltas.shape[1])
    index_map = jnp.stack(jnp.meshgrid(i, j, indexing='ij'), axis=-1)

    # Compute the pixel convergence array after applying deltas.
    convergence = jnp.rint(deltas + index_map).astype(int)

    # Pad convergence and labels arrays.
    pad_width = (((kernel_size[0] - 1) // 2, ) * 2, ((kernel_size[1] - 1) // 2, ) * 2)
    convergence = jnp.pad(convergence, (*pad_width, (0, 0)))
    labels = jnp.pad(labels, pad_width)

    # Sum pool the label values of pixels that converge at each pixel location.
    pooled_labels = vmap_sum_convergent_labels(convergence, labels, kernel_size, i, j)

    return pooled_labels


vmap_sum_pool = vmap(sum_pool, in_axes=(0, 0, None))


def _sum_convergent_labels(
        convergence: jax.Array,
        labels: jax.Array,
        kernel_size: Sequence[int],
        i: jax.Array,
        j: jax.Array
) -> jax.Array:

    """Sum pool the label values of pixels that converge at a given pixel location.

    Parameters
    ----------
    convergence : jax.Array
        Convergence array.
    labels : jax.Array
        Binary labels.
    kernel_size : Sequence[int]
        Kernel size or window size to search for labels convergence.
    i : jax.Array
        Pixel row index.
    j : jax.Array
        Pixel column index.

    Returns
    -------
    pooled_label : jax.Array
        Pooled label.
    """

    # Extract the convergence and labels arrays in the kernel.
    convergence = dynamic_slice(convergence, (i, j, 0), (*kernel_size, 2))
    labels = dynamic_slice(labels, (i, j), kernel_size)

    # Find pixel sources that converge at the given pixel location.
    sources = jnp.all(convergence == jnp.array((i, j)), axis=-1)

    # Sum pool the label values at pixel sources.
    pooled_label = jnp.sum(sources * labels)

    return pooled_label


vmap_sum_convergent_labels = vmap(
    vmap(_sum_convergent_labels, in_axes=(None, None, None, None, 0)),
    in_axes=(None, None, None, 0, None)
)


def smooth_sum_pool(
        deltas: jax.Array,
        labels: jax.Array,
        sigma: float = 0.5,
        kernel_size: Sequence[int] = (3, 3),
        epsilon: float = 1e-7
) -> jax.Array:

    """Sum pool labels using deltas.

    Unlike conventional pooling, which takes a discrete summation, each pixel's contribution to the pooled result of
    neighboring pixels is determined by an isotropic Gaussian distribution. This Gaussian is centered at the pixel's
    convergence coordinates given by `delta` and has standard deviation `sigma` in each axis. This operation is
    designed to be differentiable and thus suitable to be used within a loss function during training.

    Parameters
    ----------
    deltas : jax.Array
        Displacement vectors.
    labels : jax.Array
        Binary labels.
    sigma : jax.Array
        Standard deviation of the Gaussian distribution. Default is 0.5.
    kernel_size : Sequence[int], optional
        Kernel size or window size of the sum pooling operation. Default is (3, 3).
    epsilon : float, optional
        Small constant for numerical stability. Default is 1e-7.

    Returns
    -------
    pooled_labels : jax.Array
        Pooled labels.
    """

    # Generate an index map.
    i, j = jnp.arange(deltas.shape[0]), jnp.arange(deltas.shape[1])
    index_map = jnp.stack(jnp.meshgrid(i, j, indexing='ij'), axis=-1)

    # Compute the pixel convergence array after applying deltas.
    convergence = deltas + index_map

    # Pad convergence and labels arrays.
    pad_width = (((kernel_size[0] - 1) // 2, ) * 2, ((kernel_size[1] - 1) // 2, ) * 2)
    index_map = jnp.pad(index_map, (*pad_width, (0, 0)))
    labels = jnp.pad(labels, pad_width)

    # Compute Gaussian distributions.
    gaussians = vmap_compute_gaussian_distributions(index_map, convergence, sigma, kernel_size, epsilon, i, j)
    gaussians = jnp.pad(gaussians, (*pad_width, (0, 0), (0, 0)))

    # Distribute labels.
    pooled_labels = vmap_distribute_labels(gaussians, labels, kernel_size, i, j)

    return pooled_labels


vmap_smooth_sum_pool = vmap(smooth_sum_pool, in_axes=(0, 0, None, None, None))


def _compute_gaussian_distributions(
        index_map: jax.Array,
        convergence: jax.Array,
        sigma: float,
        kernel_size: Sequence[int],
        epsilon: float,
        i: jax.Array,
        j: jax.Array
) -> jax.Array:

    """Compute a Gaussian distribution centered at the convergence coordinates of a given pixel location.

    Parameters
    ----------
    index_map : jax.Array
        Index map.
    convergence : jax.Array
        Convergence array.
    sigma : float
        Standard deviation of the Gaussian distribution.
    kernel_size : Sequence[int]
        Kernel size or window size of the Gaussian distribution.
    epsilon : float
        Small constant for numerical stability.
    i : jax.Array
        Pixel row index.
    j : jax.Array
        Pixel column index.

    Returns
    -------
    gaussian : jax.Array
        Gaussian distribution centered at the convergence coordinates of the given pixel location.
    """

    # Extract the index map in the kernel.
    index_map = dynamic_slice(index_map, (i, j, 0), (*kernel_size, 2))

    # Compute the Gaussian distribution centered at the convergence coordinates.
    gaussian = jnp.exp(-jnp.sum((index_map - convergence) ** 2, axis=-1) / (2 * sigma ** 2))
    gaussian = gaussian / (jnp.sum(gaussian) + epsilon)

    return gaussian


vmap_compute_gaussian_distributions = vmap(
    vmap(_compute_gaussian_distributions, in_axes=(None, 0, None, None, None, None, 0)),
    in_axes=(None, 0, None, None, None, 0, None)
)


def _distribute_labels(
        distributions: jax.Array,
        labels: jax.Array,
        kernel_size: Sequence[int],
        i: jax.Array,
        j: jax.Array
) -> jax.Array:

    """Distribute the label values of nearby pixels according to a given set of distributions.

    Parameters
    ----------
    distributions : jax.Array
        Distributions array.
    labels : jax.Array
        Binary labels.
    kernel_size : Sequence[int]
        Kernel size or window size to distribute labels.
    i : jax.Array
        Pixel row index.
    j : jax.Array
        Pixel column index.

    Returns
    -------
    pooled_label : jax.Array
        Pooled label.
    """

    # Extract the distributions and labels arrays in the kernel.
    distributions = dynamic_slice(distributions, (i, j, 0, 0), (*kernel_size, *kernel_size))
    labels = dynamic_slice(labels, (i, j), kernel_size)

    # Extract weights from the distributions array.
    k, m = np.indices(kernel_size)
    weights = distributions[k, m, kernel_size[0] - 1 - k, kernel_size[1] - 1 - m]

    # Distribute the label values of nearby pixels.
    pooled_label = jnp.sum(weights * labels)

    return pooled_label


vmap_distribute_labels = vmap(
    vmap(_distribute_labels, in_axes=(None, None, None, None, 0)),
    in_axes=(None, None, None, 0, None)
)


def pad_and_stack(images: Sequence[np.ndarray]) -> np.ndarray:

    """Pad and stack images.

    Parameters
    ----------
    images : Sequence[np.ndarray]
        List of images to pad and stack.

    Returns
    -------
    stacked_images : np.ndarray
        Stacked images.
    """

    padded_images = pad(images)
    stacked_images = np.stack(padded_images)

    return stacked_images


def pad(images: Sequence[np.ndarray]) -> Sequence[np.ndarray]:

    """Pad images to the same size.

    Parameters
    ----------
    images : Sequence[np.ndarray]
        List of images to pad.

    Returns
    -------
    padded_images : Sequence[np.ndarray]
        List of padded images.
    """

    # Compute the padded image size.
    padded_size = [max([image.shape[i] for image in images]) for i in range(images[0].ndim)]

    # Pad images.
    padded_images = []
    for image in images:
        if image.shape == padded_size:
            padded_image = image
        else:
            pad_width = [(0, m - n) for m, n in zip(padded_size, image.shape)]
            padded_image = np.pad(image, pad_width)
        padded_images.append(padded_image)

    return padded_images


def remove_duplicate_coords(
        coords: np.ndarray,
        threshold: int = 1
) -> np.ndarray:

    """Remove duplicate coordinates within a distance threshold.

    Parameters
    ----------
    coords : np.ndarray
        Coordinates.
    threshold : int, optional
        Distance threshold. Default is 1.

    Returns
    -------
    new_coords : np.ndarray
        Coordinates without duplicates.
    """

    # Create a dictionary to keep track of sets of duplicate coordinates.
    sets = {}

    # Create a dictionary to keep track of the set id of each coordinate.
    coord_set_ids = {i: None for i in range(len(coords))}

    # Create a list to keep track of coordinates that have been checked.
    checked = []

    # Loop through each coordinate and check if it has already been assigned to a set.
    for i, set_id in coord_set_ids.items():

        # If the coordinate has not been assigned to a set yet.
        if set_id is None:

            # Create a new set and add the coordinate to it.
            set_id = len(sets)
            coord_set_ids[i] = set_id
            new_set = [i]
            sets[set_id] = new_set

            # Find other coordinates within the distance threshold and add them to the same set.
            matches = _match_coords(coords, i, set_id, coord_set_ids, checked, threshold)
            new_set.extend(matches)
            to_check = matches.copy()

            # Construct cluster of neighboring coordinates.
            while len(to_check) > 0:

                # Loop through new coordinates and find other coordinates within the distance threshold.
                for j in to_check:
                    matches = _match_coords(coords, j, set_id, coord_set_ids, checked, threshold)
                    to_check.remove(j)
                    new_matches = set(matches) - set(new_set)
                    new_set.extend(new_matches)
                    to_check.extend(new_matches)

                to_check = list(set(to_check) - set(checked))

    # Compute the mean of coordinates in each set.
    new_coords = []
    for s in sets.values():
        old_coords = coords[list(s)]
        new_coords.append(np.mean(old_coords, axis=0))
    new_coords = np.array(new_coords)

    return new_coords


def _match_coords(
        coords: np.ndarray,
        i: int,
        set_id: int,
        coord_set_ids: Dict[int, int],
        checked: List[int],
        threshold: int
) -> List[int]:

    """Match coordinates within a distance threshold and assign them to the same set.

    Parameters
    ----------
    coords : np.ndarray
        Coordinates.
    i : int
        Coordinate index.
    set_id : int
        Set id.
    coord_set_ids : Dict[int, int]
        Dictionary to keep track of the set id of each coordinate.
    checked : List[int]
        List to keep track of coordinates that have been checked.
    threshold : int
        Distance threshold.

    Returns
    -------
    matches : List[int]
        List of matched coordinates.
    """

    # Mean the distances between the ith coordinate and all other coordinates.
    distances = np.sqrt(np.sum((coords[i] - coords) ** 2, axis=1))

    # Find coordinates within the distance threshold.
    matches = list(np.where(distances < threshold)[0])

    # Remove the ith coordinate itself from the matches.
    matches.remove(i)

    # Assign the same set id to all matched coordinates.
    for match in matches:
        coord_set_ids[match] = set_id

    # Add the ith coordinate to the list of checked coordinates.
    checked.append(i)

    return matches


def snap_coords(
        coords: np.ndarray,
        image: np.ndarray,
        window_size: int = 3
) -> np.ndarray:

    """Snap each coordinate to the local maxima of an image in a window around the coordinate.

    Parameters
    ----------
    coords : np.ndarray
        Coordinates.
    image : np.ndarray
        Image.
    window_size : int, optional
        Window size. Must be an odd integer. Default is 3.

    Returns
    -------
    snapped_coords : np.ndarray
        Snapped coordinates.

    Raises
    ------
    ValueError
        If `window_size` is not an odd integer.
    """

    # Check the window size.
    if window_size % 2 == 0:
        raise ValueError("window_size must be an odd integer.")

    # Remove coordinates outside the image.
    coords = coords[(coords[:, 0] > -0.5) & (coords[:, 1] > -0.5) &
                    (coords[:, 0] < image.shape[0] - 0.5) & (coords[:, 1] < image.shape[1] - 0.5)]

    # Create a list to store the snapped coordinates.
    snapped_coords = []

    # Pad the image.
    delta = round((window_size - 1) / 2)
    image = np.pad(image, ((delta, delta), (delta, delta)))

    for coord in coords:

        # Crop the image around the coordinate.
        index = np.rint(coord).astype(int)
        cropped_image = image[index[0]: index[0] + window_size, index[1]: index[1] + window_size]

        # Find the local maximum.
        maximum = np.where(cropped_image == np.max(cropped_image))
        maximum = np.array([maximum[0][0], maximum[1][0]])

        # Snap the coordinate to the local maximum.
        offset = maximum - delta
        snapped_coords.append(index + offset)

    snapped_coords = np.array(snapped_coords)

    return snapped_coords


def fit_coords(
        coords: np.ndarray,
        image: np.ndarray,
        window_size: int = 3,
        max_gaussian_amplitude: float = 2.0,
        keep_failed_fits: bool = False
) -> np.ndarray:

    """Fit a Gaussian to the image in a window around each coordinate and return the center of the Gaussian.

    Parameters
    ----------
    coords : np.ndarray
        Coordinates.
    image : np.ndarray
        Image.
    window_size : int, optional
        Window size. Must be an odd integer. Default is 3.
    max_gaussian_amplitude : float, optional
        Maximum amplitude of the Gaussian used to fit the normalized image within a window. Default is 2.0.
    keep_failed_fits : bool, optional
        Whether to keep the original coordinates when the Gaussian fit fails. Default is False.

    Returns
    -------
    fitted_coords : np.ndarray
        Fitted coordinates.

    Raises
    ------
    ValueError
        If `window_size` is not an odd integer.
    """

    # Check the window size.
    if window_size % 2 == 0:
        raise ValueError("window_size must be an odd integer.")

    # Remove coordinates outside the image.
    coords = coords[(coords[:, 0] > -0.5) & (coords[:, 1] > -0.5) &
                    (coords[:, 0] < image.shape[0] - 0.5) & (coords[:, 1] < image.shape[1] - 0.5)]

    # Create a list to store the fitted coordinates.
    fitted_coords = []

    # Pad the image.
    delta = round((window_size - 1) / 2)
    image = np.pad(image, ((delta, delta), (delta, delta)), mode='reflect')

    # Create a meshgrid for the Gaussian fit.
    x = np.arange(-delta, delta + 1)
    x, y = np.meshgrid(x, x)

    # Define initial parameters and bounds for the Gaussian fit.
    p0 = (1, 0, 0, delta, delta, 0, 0)
    bounds = ((0, -delta, -delta, 0, 0, -np.pi, 0),
              (max_gaussian_amplitude, delta, delta, np.inf, np.inf, np.pi, max_gaussian_amplitude))

    for coord in coords:

        # Crop the image around the coordinate.
        index = np.rint(coord).astype(int)
        cropped_image = image[index[0]: index[0] + window_size, index[1]: index[1] + window_size].ravel()

        # Normalize the cropped image.
        image_min = np.min(cropped_image)
        image_max = np.max(cropped_image)
        cropped_image = (cropped_image - image_min) / (image_max - image_min + 1e-7)

        # Fit a Gaussian to the cropped image.
        try:
            popt, pcov = optimize.curve_fit(_gaussian, (x, y), cropped_image, p0=p0, bounds=bounds)
            fitted_coords.append(index + np.array([popt[2], popt[1]]))
        except RuntimeError:
            if keep_failed_fits:
                fitted_coords.append(coord)

    fitted_coords = np.array(fitted_coords)

    return fitted_coords


@njit
def _gaussian(
        xy: Tuple[np.ndarray, np.ndarray],
        amplitude: float,
        x0: float,
        y0: float,
        sigma_x: float,
        sigma_y: float,
        theta: float,
        offset: float
) -> np.ndarray:

    """2D Gaussian function.

    Parameters
    ----------
    xy : Tuple[np.ndarray, np.ndarray]
        x and y values at which to evaluate the Gaussian.
    amplitude : float
        Amplitude of the Gaussian.
    x0 : float
        x-coordinate of the center of the Gaussian.
    y0 : float
        y-coordinate of the center of the Gaussian.
    sigma_x : float
        Standard deviation of the Gaussian in the x-direction.
    sigma_y : float
        Standard deviation of the Gaussian in the y-direction.
    theta : float
        Rotation angle of the Gaussian.
    offset : float
        Offset of the Gaussian in the z-direction.

    Returns
    -------
    g : np.ndarray
        Gaussian at the given x and y values.
    """

    x, y = xy
    a = (np.cos(theta) ** 2) / (2 * sigma_x ** 2) + (np.sin(theta) ** 2) / (2 * sigma_y ** 2)
    b = -(np.sin(2 * theta)) / (4 * sigma_x ** 2) + (np.sin(2 * theta))/(4 * sigma_y ** 2)
    c = (np.sin(theta) ** 2)/(2 * sigma_x ** 2) + (np.cos(theta) ** 2)/(2 * sigma_y ** 2)
    g = offset + amplitude * np.exp(-(a * ((x - x0) ** 2) + 2 * b * (x - x0) * (y - y0) + c * ((y - y0) ** 2)))
    g = g.ravel()

    return g



import numpy as np
import pandas as pd

from scipy import integrate, optimize, spatial
from typing import Dict, Iterable, Optional, Sequence, Tuple, Union


def compute_metrics(
        coords_pred: np.ndarray,
        coords: np.ndarray,
        evaluation_metrics: Union[str, Sequence[str]] = 'f1',
        distance_metric: str = 'euclidean',
        distance_thresholds: Union[float, Iterable] = 1.0,
        epsilon: float = 1e-7,
        return_df: bool = False
) -> Union[Dict[str, float], Tuple[Dict[str, float], pd.DataFrame]]:

    """Compute evaluation metrics from predicted and ground truth spot coordinates. Adapted from deepBlink.

    Parameters
    ----------
    coords_pred : np.ndarray
        Predicted spot coordinates.
    coords : np.ndarray
        Ground truth spot coordinates.
    evaluation_metrics : Union[str, Sequence[str]], optional
        Evaluation metric or a list of evaluation metrics. Supported evaluation metrics are 'f1', 'precision', 'recall',
        'tp', 'fp', and 'fn'. Default is 'f1'.
    distance_metric : str, optional
        Distance metric used to compute the distances between predicted and ground truth spot coordinates. Default is
        'euclidean'.
    distance_thresholds : Union[float, Iterable], optional
        Distance threshold or a list of distance thresholds for matching predicted and ground truth spot coordinates.
        Default is 1.0.
    epsilon : float, optional
        Small constant for numerical stability. Default is 1e-7.
    return_df : bool, optional
        Whether to return a DataFrame with additional information. Default is False.

    Returns
    -------
    integral_metrics : Dict[str, float]
        Dictionary of aggregate metrics integrated over the distance thresholds.
    df : pd.DataFrame, optional
        DataFrame with additional information. Only returned if `return_df` is True.

    References
    ----------
    .. [1] Eichenberger, Bastian, et al. "BBQuercus/deepBlink: Threshold independent detection and localization of
           diffraction-limited spots." GitHub, https://github.com/BBQuercus/deepBlink.
    """

    # Compute the distance matrix between predicted and ground truth spot coordinates.
    distance_matrix = spatial.distance.cdist(coords, coords_pred, metric=distance_metric)

    # Convert `evaluation_metrics` to a tuple if necessary.
    if isinstance(evaluation_metrics, str):
        evaluation_metrics = (evaluation_metrics,)

    # Convert `distance_thresholds` to a tuple if necessary.
    if not isinstance(distance_thresholds, Iterable):
        distance_thresholds = (distance_thresholds, )

    # Create empty lists.
    metrics_list = []
    deltas_list = []
    offsets_list = []

    if return_df:

        # Loop over distance thresholds.
        for distance_threshold in distance_thresholds:

            # Compute evaluation metrics.
            metrics, coords_matches = _compute_metrics(distance_matrix, evaluation_metrics, distance_threshold, epsilon)

            # Compute deltas and offsets.
            deltas = coords[coords_matches[0]] - coords_pred[coords_matches[1]]
            offsets = distance_matrix[coords_matches]

            # Add evaluation metrics, deltas, and offsets.
            metrics_list.append(metrics)
            deltas_list.append(deltas)
            offsets_list.append(offsets)

    else:

        # Loop over distance thresholds.
        for distance_threshold in distance_thresholds:

            # Compute evaluation metrics.
            metrics, _ = _compute_metrics(distance_matrix, evaluation_metrics, distance_threshold, epsilon)

            # Add evaluation metrics.
            metrics_list.append(metrics)

    # Create an empty dictionary.
    integral_metrics = {}

    # Integrate evaluation metrics over distance thresholds.
    if len(distance_thresholds) > 1:
        for evaluation_metric in evaluation_metrics:
            integral_metrics[evaluation_metric] = (
                    integrate.trapz([metrics[evaluation_metric] for metrics in metrics_list], distance_thresholds) /
                    np.ptp(distance_thresholds))
    else:
        integral_metrics = metrics_list[0]

    if return_df:

        # Create a DataFrame with deltas, offsets, and evaluation metrics for each distance threshold.
        df = pd.DataFrame(
            {'thresholds': distance_thresholds, 'deltas': deltas_list, 'offsets': offsets_list} |
            {k: [metrics[k] for metrics in metrics_list] for k in evaluation_metrics}
        )

        return integral_metrics, df

    else:

        return integral_metrics


def _compute_metrics(
        distance_matrix: np.ndarray,
        evaluation_metrics: Union[str, Sequence[str]],
        distance_threshold: float,
        epsilon: float,
) -> Tuple[Dict[str, float], Tuple[np.ndarray, np.ndarray]]:

    """Compute evaluation metrics from a given distance matrix.

    Parameters
    ----------
    distance_matrix : np.ndarray
        Distance matrix.
    evaluation_metrics : Union[str, Sequence[str]]
        Evaluation metric or a list of evaluation metrics. Supported evaluation metrics are 'f1', 'precision', 'recall',
        'tp', 'fp', and 'fn'.
    distance_threshold : float
        Distance threshold for matching predicted and ground truth spot coordinates.
    epsilon : float
        Small constant for numerical stability.
    """

    # Match predicted and ground truth spot coordinates.
    coords_matches = linear_sum_assignment(distance_matrix, distance_threshold)

    # Compute the number of true positives, false positives, and false negatives.
    tp = len(coords_matches[0])
    fp = distance_matrix.shape[1] - tp
    fn = distance_matrix.shape[0] - tp

    # Compute evaluation metrics.
    metrics = {}
    if 'f1' in evaluation_metrics:
        metrics['f1'] = 2 * tp / (2 * tp + fp + fn + epsilon)
    if 'precision' in evaluation_metrics:
        metrics['precision'] = tp / (tp + fp + epsilon)
    if 'recall' in evaluation_metrics:
        metrics['recall'] = tp / (tp + fn + epsilon)
    if 'tp' in evaluation_metrics:
        metrics['tp'] = tp
    if 'fp' in evaluation_metrics:
        metrics['fp'] = fp
    if 'fn' in evaluation_metrics:
        metrics['fn'] = fn

    return metrics, coords_matches


def linear_sum_assignment(
        cost_matrix: np.ndarray,
        threshold: Optional[float] = None
) -> Tuple[np.ndarray, np.ndarray]:

    """Solve the linear sum assignment problem for a given cost matrix.

    Parameters
    ----------
    cost_matrix : np.ndarray
        Cost matrix.
    threshold : float, optional
        Threshold for matching. Default is None.

    Returns
    -------
    matches : Tuple[np.ndarray, np.ndarray]
        Tuple of row and column indices of the matches.
    """

    if threshold is not None:
        cost_matrix = np.where(cost_matrix > threshold, np.max(cost_matrix), cost_matrix)
    matches = optimize.linear_sum_assignment(cost_matrix)
    below_threshold = [i for i, match in enumerate(zip(*matches)) if cost_matrix[match] <= threshold]
    matches = (matches[0][below_threshold], matches[1][below_threshold])

    return matches



DEFAULT_BLOCKS_ARGS = {
    'efficientnetv2-s': [
        {
            'kernel_size': (3, 3),
            'num_repeat': 2,
            'features_in': 24,
            'features_out': 24,
            'expand_ratio': 1,
            'se_ratio': 0.0,
            'strides': 1,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 4,
            'features_in': 24,
            'features_out': 48,
            'expand_ratio': 4,
            'se_ratio': 0.0,
            'strides': 2,
            'conv_type': 1,
        },
        {
            'conv_type': 1,
            'expand_ratio': 4,
            'features_in': 48,
            'kernel_size': (3, 3),
            'num_repeat': 4,
            'features_out': 64,
            'se_ratio': 0,
            'strides': 2,
        },
        {
            'conv_type': 0,
            'expand_ratio': 4,
            'features_in': 64,
            'kernel_size': (3, 3),
            'num_repeat': 6,
            'features_out': 128,
            'se_ratio': 0.25,
            'strides': 2,
        },
        {
            'conv_type': 0,
            'expand_ratio': 6,
            'features_in': 128,
            'kernel_size': (3, 3),
            'num_repeat': 9,
            'features_out': 160,
            'se_ratio': 0.25,
            'strides': 1,
        },
        {
            'conv_type': 0,
            'expand_ratio': 6,
            'features_in': 160,
            'kernel_size': (3, 3),
            'num_repeat': 15,
            'features_out': 256,
            'se_ratio': 0.25,
            'strides': 2,
        },
    ],
    'efficientnetv2-m': [
        {
            'kernel_size': (3, 3),
            'num_repeat': 3,
            'features_in': 24,
            'features_out': 24,
            'expand_ratio': 1,
            'se_ratio': 0,
            'strides': 1,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 5,
            'features_in': 24,
            'features_out': 48,
            'expand_ratio': 4,
            'se_ratio': 0,
            'strides': 2,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 5,
            'features_in': 48,
            'features_out': 80,
            'expand_ratio': 4,
            'se_ratio': 0,
            'strides': 2,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 7,
            'features_in': 80,
            'features_out': 160,
            'expand_ratio': 4,
            'se_ratio': 0.25,
            'strides': 2,
            'conv_type': 0,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 14,
            'features_in': 160,
            'features_out': 176,
            'expand_ratio': 6,
            'se_ratio': 0.25,
            'strides': 1,
            'conv_type': 0,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 18,
            'features_in': 176,
            'features_out': 304,
            'expand_ratio': 6,
            'se_ratio': 0.25,
            'strides': 2,
            'conv_type': 0,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 5,
            'features_in': 304,
            'features_out': 512,
            'expand_ratio': 6,
            'se_ratio': 0.25,
            'strides': 1,
            'conv_type': 0,
        },
    ],
    'efficientnetv2-l': [
        {
            'kernel_size': (3, 3),
            'num_repeat': 4,
            'features_in': 32,
            'features_out': 32,
            'expand_ratio': 1,
            'se_ratio': 0,
            'strides': 1,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 7,
            'features_in': 32,
            'features_out': 64,
            'expand_ratio': 4,
            'se_ratio': 0,
            'strides': 2,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 7,
            'features_in': 64,
            'features_out': 96,
            'expand_ratio': 4,
            'se_ratio': 0,
            'strides': 2,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 10,
            'features_in': 96,
            'features_out': 192,
            'expand_ratio': 4,
            'se_ratio': 0.25,
            'strides': 2,
            'conv_type': 0,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 19,
            'features_in': 192,
            'features_out': 224,
            'expand_ratio': 6,
            'se_ratio': 0.25,
            'strides': 1,
            'conv_type': 0,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 25,
            'features_in': 224,
            'features_out': 384,
            'expand_ratio': 6,
            'se_ratio': 0.25,
            'strides': 2,
            'conv_type': 0,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 7,
            'features_in': 384,
            'features_out': 640,
            'expand_ratio': 6,
            'se_ratio': 0.25,
            'strides': 1,
            'conv_type': 0,
        },
    ],
    'efficientnetv2-b0': [
        {
            'kernel_size': (3, 3),
            'num_repeat': 1,
            'features_in': 32,
            'features_out': 16,
            'expand_ratio': 1,
            'se_ratio': 0,
            'strides': 1,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 2,
            'features_in': 16,
            'features_out': 32,
            'expand_ratio': 4,
            'se_ratio': 0,
            'strides': 2,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 2,
            'features_in': 32,
            'features_out': 48,
            'expand_ratio': 4,
            'se_ratio': 0,
            'strides': 2,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 3,
            'features_in': 48,
            'features_out': 96,
            'expand_ratio': 4,
            'se_ratio': 0.25,
            'strides': 2,
            'conv_type': 0,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 5,
            'features_in': 96,
            'features_out': 112,
            'expand_ratio': 6,
            'se_ratio': 0.25,
            'strides': 1,
            'conv_type': 0,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 8,
            'features_in': 112,
            'features_out': 192,
            'expand_ratio': 6,
            'se_ratio': 0.25,
            'strides': 2,
            'conv_type': 0,
        },
    ],
    'efficientnetv2-b1': [
        {
            'kernel_size': (3, 3),
            'num_repeat': 1,
            'features_in': 32,
            'features_out': 16,
            'expand_ratio': 1,
            'se_ratio': 0,
            'strides': 1,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 2,
            'features_in': 16,
            'features_out': 32,
            'expand_ratio': 4,
            'se_ratio': 0,
            'strides': 2,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 2,
            'features_in': 32,
            'features_out': 48,
            'expand_ratio': 4,
            'se_ratio': 0,
            'strides': 2,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 3,
            'features_in': 48,
            'features_out': 96,
            'expand_ratio': 4,
            'se_ratio': 0.25,
            'strides': 2,
            'conv_type': 0,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 5,
            'features_in': 96,
            'features_out': 112,
            'expand_ratio': 6,
            'se_ratio': 0.25,
            'strides': 1,
            'conv_type': 0,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 8,
            'features_in': 112,
            'features_out': 192,
            'expand_ratio': 6,
            'se_ratio': 0.25,
            'strides': 2,
            'conv_type': 0,
        },
    ],
    'efficientnetv2-b2': [
        {
            'kernel_size': (3, 3),
            'num_repeat': 1,
            'features_in': 32,
            'features_out': 16,
            'expand_ratio': 1,
            'se_ratio': 0,
            'strides': 1,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 2,
            'features_in': 16,
            'features_out': 32,
            'expand_ratio': 4,
            'se_ratio': 0,
            'strides': 2,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 2,
            'features_in': 32,
            'features_out': 48,
            'expand_ratio': 4,
            'se_ratio': 0,
            'strides': 2,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 3,
            'features_in': 48,
            'features_out': 96,
            'expand_ratio': 4,
            'se_ratio': 0.25,
            'strides': 2,
            'conv_type': 0,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 5,
            'features_in': 96,
            'features_out': 112,
            'expand_ratio': 6,
            'se_ratio': 0.25,
            'strides': 1,
            'conv_type': 0,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 8,
            'features_in': 112,
            'features_out': 192,
            'expand_ratio': 6,
            'se_ratio': 0.25,
            'strides': 2,
            'conv_type': 0,
        },
    ],
    'efficientnetv2-b3': [
        {
            'kernel_size': (3, 3),
            'num_repeat': 1,
            'features_in': 32,
            'features_out': 16,
            'expand_ratio': 1,
            'se_ratio': 0,
            'strides': 1,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 2,
            'features_in': 16,
            'features_out': 32,
            'expand_ratio': 4,
            'se_ratio': 0,
            'strides': 2,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 2,
            'features_in': 32,
            'features_out': 48,
            'expand_ratio': 4,
            'se_ratio': 0,
            'strides': 2,
            'conv_type': 1,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 3,
            'features_in': 48,
            'features_out': 96,
            'expand_ratio': 4,
            'se_ratio': 0.25,
            'strides': 2,
            'conv_type': 0,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 5,
            'features_in': 96,
            'features_out': 112,
            'expand_ratio': 6,
            'se_ratio': 0.25,
            'strides': 1,
            'conv_type': 0,
        },
        {
            'kernel_size': (3, 3),
            'num_repeat': 8,
            'features_in': 112,
            'features_out': 192,
            'expand_ratio': 6,
            'se_ratio': 0.25,
            'strides': 2,
            'conv_type': 0,
        },
    ],
}



from piscis.core import Piscis


def _initialize_jax_cache() -> None:

    """Initialize the JAX compilation cache."""

    from jax._src import compilation_cache
    from piscis.paths import CACHE_DIR

    # Set the compilation cache directory.
    compilation_cache.set_cache_dir(str(CACHE_DIR))


_initialize_jax_cache()
del _initialize_jax_cache



import dask.array as da
import deeptile
import jax
import jax.numpy as jnp
import numpy as np
import xarray as xr

from deeptile import lift, Output
from deeptile.core.data import Tiled
from deeptile.extensions import stitch
from flax import serialization
from functools import partial
from jax import jit
from jax.lib import xla_bridge
from skimage.transform import resize
from typing import Dict, Optional, Sequence, Tuple, Union

from piscis import utils
from piscis.downloads import download_pretrained_model
from piscis.models.spots import round_input_size, SpotsModel
from piscis.paths import MODELS_DIR


class Piscis:

    """Class for running the Piscis algorithm.

    Attributes
    ----------
    model_name : str
        Model name.
    batch_size : int
        Batch size for the CNN.
    variables : Dict
        Model variables.
    adjustment : str
        Adjustment type applied to images during preprocessing.
    input_size : Tuple[int, int]
        Input size for the CNN.
    dilation_iterations : int
        Number of iterations used to dilate ground truth labels during training.
    apply : Callable
        Model apply function.
    """

    def __init__(
            self,
            model_name: str = '20230905',
            batch_size: int = 4,
            input_size: Optional[Tuple[int, int]] = None
    ) -> None:

        """Initialize a Piscis object and compile the model.

        Parameters
        ----------
        model_name : str, optional
            Model name. Default is '20230905'.
        batch_size : int, optional
            Batch size for the CNN. Default is 4.
        input_size : Optional[Tuple[int, int]], optional
            Input size for the CNN. If None, it is obtained from the model dictionary. Default is None.
        """

        # Set the batch size to 1 if running on CPU.
        if xla_bridge.get_backend().platform == 'cpu':
            batch_size = 1

        # Load the model.
        self.model_name = model_name
        self.batch_size = batch_size
        model_path = MODELS_DIR / model_name
        if not model_path.is_file():
            download_pretrained_model(model_name)
        with open(MODELS_DIR / model_name, 'rb') as f_model:
            model_dict = serialization.from_bytes(target=None, encoded_bytes=f_model.read())
            self.variables = model_dict['variables']
            self.adjustment = model_dict['adjustment']
            if input_size is None:
                input_size = model_dict['input_size']
                input_size = (input_size['0'], input_size['1'])
            else:
                input_size = round_input_size(input_size)
            self.input_size = input_size
            self.dilation_iterations = model_dict['dilation_iterations']
            self.channels = model_dict.get('channels', 1)

        # Define the model apply function.
        kernel_size = (2 * self.dilation_iterations + 1, ) * 2
        self.apply = partial(apply, variables=self.variables, kernel_size=kernel_size)

    def predict(
            self,
            x: Union[np.ndarray, da.Array],
            stack: bool = False,
            scale: float = 1.0,
            threshold: float = 1.0,
            min_distance: int = 1,
            intermediates: bool = False
    ) -> Union[Tuple[np.ndarray, xr.DataArray], np.ndarray]:

        """Predict spots in an image or stack of images.

        Parameters
        ----------
        x : np.ndarray or da.Array
            Image or stack of images.
        stack : bool, optional
            Whether `x` is a stack of images. Default is False.
        scale : float, optional
            Scale factor for rescaling `x`. Default is 1.
        threshold : float, optional
            Spot detection threshold. Can be interpreted as the minimum number of fully confident pixels necessary to
            identify a spot. Default is 1.0.
        min_distance : int, optional
            Minimum distance between spots. Default is 1.
        intermediates : bool, optional
            Whether to return intermediate feature maps. Default is False.

        Returns
        -------
        coords : np.ndarray
            Predicted spot coordinates.
        y : np.ndarray, optional
            Intermediate feature maps. Only returned if `intermediates` is True.
        """

        # Preprocess the input.
        x, batch_axis, x_stats = self._preprocess(x, stack)

        # Create a DeepTile object and get tiles.
        dt = deeptile.load(x, link_data=False, dask=True)
        tile_size = (round(self.input_size[0] / scale), round(self.input_size[1] / scale))
        scales = (np.array([self.input_size]) - 1) / (np.array(tile_size) - 1)
        tiles = dt.get_tiles(tile_size=tile_size, overlap=(0.1, 0.1)).pad(mode='symmetric')

        # Adjust tiles if necessary.
        if x_stats is not None:
            if self.adjustment == 'normalize':
                x_min, x_max = x_stats
                tiles = lift(lambda t: (t - x_min) / (x_max - x_min + 1e-7))(tiles)
            elif self.adjustment == 'standardize':
                x_mean, x_std = x_stats
                tiles = lift(lambda t: (t - x_mean) / (x_std + 1e-7))(tiles)

        # Predict spots in tiles and stitch.
        if stack and batch_axis:

            batch_axis_len = x.shape[0]
            stack_axis_len = x.shape[-4]
            coords = np.empty(batch_axis_len, dtype=object)
            if intermediates:
                y = []
            else:
                y = None
            for i in range(batch_axis_len):
                output = self._predict_stack(tiles.s[i], stack_axis_len, scales, threshold, min_distance, intermediates)
                if intermediates:
                    c = stitch.stitch_coords(output[0])
                    y.append(np.asarray(stitch.stitch_image(output[1], blend=False)))
                else:
                    c = stitch.stitch_coords(output)
                coords[i] = np.asarray(c)
            if intermediates:
                y = np.stack(y)

        else:

            if stack:
                stack_axis_len = x.shape[-4]
                output = self._predict_stack(tiles, stack_axis_len, scales, threshold, min_distance, intermediates)
            else:
                output = lift(self._predict_plane, vectorized=True, batch_axis=batch_axis, pad_final_batch=True,
                              batch_size=self.batch_size)(tiles, scales, threshold, min_distance, intermediates)
            if intermediates:
                coords = stitch.stitch_coords(output[0])
                y = np.asarray(stitch.stitch_image(output[1], blend=False))
            else:
                coords = stitch.stitch_coords(output)
                y = None
            coords = np.asarray(coords)

        if intermediates:

            dims = tuple(dim for cond, dim in ((batch_axis, 'n'), (stack, 'z')) if cond) + ('c', 'y', 'x')
            y = xr.DataArray(y, dims=dims)

            return coords, y

        else:

            return coords

    def _predict_stack(
            self,
            tiles: Tiled,
            stack_axis_len: int,
            scales: np.ndarray,
            threshold: float,
            min_distance: int,
            intermediates: bool
    ) -> Union[Tuple[Output, np.ndarray], Output]:

        """Predict spots in a stack of images.

        Parameters
        ----------
        tiles : Tiled
            Tiles of images.
        stack_axis_len : int
            Length of the stack axis.
        scales : np.ndarray
            Scales for rescaling tiles.
        threshold : float
            Spot detection threshold. Can be interpreted as the minimum number of fully confident pixels necessary to
            identify a spot.
        min_distance : int
            Minimum distance between spots.
        intermediates : bool
            Whether to return intermediate feature maps.

        Returns
        -------
        coords : Output
            Predicted spot coordinates.
        y : np.ndarray, optional
            Intermediate feature maps. Only returned if `intermediates` is True.
        """

        # Lift process and postprocess functions.
        process_stack = lift(partial(self._process, intermediates=intermediates),
                             vectorized=True, batch_axis=True, pad_final_batch=True, batch_size=self.batch_size)
        postprocess_stack = lift(self._postprocess_stack, vectorized=False, batch_axis=False)

        # Initialize the lifted process function.
        carry, process_variables = process_stack.init(tiles)
        coords = None
        postprocess_variables = None
        i_max = tiles.shape[0]
        mod = 0
        i = 0
        j = 0

        for _ in range(process_variables['n_steps']):

            # Apply the lifted process function.
            carry, process_variables = process_stack.apply(carry, process_variables)

            # Initialize the lifted postprocess function if necessary.
            if postprocess_variables is None:
                coords, postprocess_variables = \
                    postprocess_stack.init(carry[0], carry[1], scales, threshold, min_distance)

            # Apply the lifted postprocess function.
            mod = mod + self.batch_size
            while (mod >= stack_axis_len) and (i < i_max):
                coords, postprocess_variables = postprocess_stack.apply(coords, postprocess_variables)
                if not intermediates:
                    carry[0][i, j] = None
                    carry[1][i, j] = None
                mod = mod - stack_axis_len
                if j == coords.shape[1] - 1:
                    i = i + 1
                    j = 0
                else:
                    j = j + 1

        if intermediates:
            y = np.concatenate((carry[0],
                                np.expand_dims(carry[2], axis=-1), np.expand_dims(carry[1], axis=-1)), axis=-1)
            y = np.moveaxis(y, -1, 1)
            return coords, y
        else:
            return coords

    def _predict_plane(
            self,
            tiles: Tiled,
            scales: np.ndarray,
            threshold: float,
            min_distance: int,
            intermediates: bool
    ) -> Union[Tuple[Output, np.ndarray], Output]:

        """Predict spots in a single plane.

        Parameters
        ----------
        tiles : Tiled
            Tiles of images.
        scales : np.ndarray
            Scales for rescaling tiles.
        threshold : float
            Spot detection threshold. Can be interpreted as the minimum number of fully confident pixels necessary to
            identify a spot.
        min_distance : int
            Minimum distance between spots.
        intermediates : bool
            Whether to return intermediate feature maps.

        Returns
        -------
        coords : Output
            Predicted spot coordinates.
        y : np.ndarray, optional
            Intermediate feature maps. Only returned if `intermediates` is True.
        """

        # Process tiles.
        if intermediates:
            deltas, pooled_labels, labels = self._process(tiles, intermediates=intermediates)
        else:
            deltas, pooled_labels = self._process(tiles, intermediates=intermediates)
            labels = None

        # Postprocess tiles.
        coords = np.empty(len(deltas), dtype=object)
        for i, (d, pl) in enumerate(zip(deltas, pooled_labels)):
            coords[i] = utils.compute_spot_coordinates(d, pl, threshold=threshold, min_distance=min_distance)
            coords[i][:, -2:] = coords[i][:, -2:] / scales
        coords = Output(coords, isimage=False, stackable=False, tile_scales=(1.0, 1.0))

        if intermediates:
            y = np.concatenate(
                (deltas, np.expand_dims(labels, axis=-1), np.expand_dims(pooled_labels, axis=-1)), axis=-1
            )
            y = np.moveaxis(y, -1, 1)
            return coords, y
        else:
            return coords

    def _preprocess(
            self,
            x: Union[np.ndarray, da.Array],
            stack: bool
    ) -> Tuple[da.Array, bool, Optional[Tuple[np.ndarray, np.ndarray]]]:

        """Preprocess the input.

        Parameters
        ----------
        x : np.ndarray or dask.array.Array
            Image or stack of images.
        stack : bool
            Whether `x` is a stack of images.

        Returns
        -------
        x : Union[np.ndarray, da.Array]
            Preprocessed image or stack of images.
        batch_axis : bool
            Whether `x` has a batch axis.
        x_stats : Optional[Tuple[np.ndarray, np.ndarray]]
            Statistics of `x` across each plane used for image adjustment.

        Raises
        ------
        ValueError
            If `x` does not have the correct dimensions.
        ValueError
            If `x` does not have the correct number of channels.
        """

        # Check the channel axis.
        if self.channels == 1:
            x = np.expand_dims(x, axis=-3)
        else:
            if x.shape[-3] != self.channels:
                raise ValueError("The input does not have the correct number of channels.")

        # Get the number of input dimensions.
        ndim = x.ndim

        # Check the number of dimensions.
        if ((ndim == 5) and stack) or ((ndim == 4) and (not stack)):
            batch_axis = True
        elif ((ndim == 3) and (not stack)) or ((ndim == 4) and stack):
            batch_axis = False
        else:
            raise ValueError("The input does not have the correct dimensions.")

        # Convert the input to a Dask array if necessary.
        if isinstance(x, xr.DataArray):
            x = x.data
        if not isinstance(x, da.Array):
            x = da.from_array(x)

        # Standardize the input if necessary.
        if self.adjustment == 'normalize':
            x_min = x.min(axis=(-2, -1), keepdims=True).compute()
            x_max = x.max(axis=(-2, -1), keepdims=True).compute()
            x_stats = (x_min, x_max)
        elif self.adjustment == 'standardize':
            x_mean = x.mean(axis=(-2, -1), keepdims=True).compute()
            x_std = x.std(axis=(-2, -1), keepdims=True).compute()
            x_stats = (x_mean, x_std)
        else:
            x_stats = None

        return x, batch_axis, x_stats

    def _process(
            self,
            tiles: Tiled,
            intermediates: bool
    ) -> Union[Tuple[np.ndarray, np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray]]:

        """Process tiles.

        Parameters
        ----------
        tiles : Tiled
            Tiles of images.
        intermediates : bool
            Whether to keep intermediate feature maps.

        Returns
        -------
        deltas : np.ndarray
            Predicted displacement vectors.
        pooled_labels : np.ndarray
            Predicted pooled labels.
        labels : np.ndarray, optional
            Predicted binary labels. Only returned if `intermediates` is True.
        """

        # Compute Dask arrays if necessary.
        tiles = tiles.compute()

        # Resize tiles if necessary.
        if tiles.shape[-2:] != self.input_size:
            tiles = resize(tiles, (*tiles.shape[:-2], *self.input_size), preserve_range=True)

        # Run the jitted model apply function on tiles.
        if intermediates:
            deltas, pooled_labels, labels = self.apply(tiles)
            deltas = np.asarray(deltas)
            pooled_labels = np.asarray(pooled_labels)
            labels = np.asarray(labels)
            return deltas, pooled_labels, labels
        else:
            deltas, pooled_labels, _ = self.apply(tiles)
            deltas = np.asarray(deltas)
            pooled_labels = np.asarray(pooled_labels)
            return deltas, pooled_labels

    @staticmethod
    def _postprocess_stack(
            deltas: np.ndarray,
            pooled_labels: np.ndarray,
            scales: np.ndarray,
            threshold: float,
            min_distance: int
    ) -> Output:

        """Postprocess tiles.

        Parameters
        ----------
        deltas : np.ndarray
            Predicted displacement vectors.
        pooled_labels : np.ndarray
            Predicted pooled labels.
        scales : np.ndarray
            Scales for rescaling tiles.
        threshold : float
            Spot detection threshold. Can be interpreted as the minimum number of fully confident pixels necessary to
            identify a spot.
        min_distance : int
            Minimum distance between spots.

        Returns
        -------
        coords : Output
            Predicted spot coordinates.
        """

        # Compute spot coordinates.
        coords = utils.compute_spot_coordinates(deltas, pooled_labels, threshold=threshold, min_distance=min_distance)

        # Rescale coordinates.
        coords[:, -2:] = coords[:, -2:] / scales
        coords = Output(coords, isimage=False, stackable=False, tile_scales=(1.0, 1.0))

        return coords


@partial(jit, static_argnums=2)
def apply(
        x: jax.Array,
        variables: Dict,
        kernel_size: Sequence[int]
) -> Tuple[jax.Array, jax.Array, jax.Array]:

    """Apply SpotsModel to a batch images.

    x : jax.Array
        Batch of images.
    variables : Dict
        Model variables.
    kernel_size : Sequence[int]
        Kernel size or window size of the sum pooling operation.

    Returns
    -------
    deltas : jax.Array
        Predicted displacements vectors.
    pooled_labels : jax.Array
        Predicted pooled labels.
    labels : jax.Array
        Predicted binary labels.
    """

    x = jnp.moveaxis(x, -3, -1)
    deltas, labels = SpotsModel().apply(variables, x, False)
    labels = labels[:, :, :, 0]
    pooled_labels = utils.vmap_sum_pool(deltas, labels, kernel_size)

    return deltas, pooled_labels, labels


def adjust_parameters(
        y: xr.DataArray,
        threshold: float = 1.0,
        min_distance: int = 1
) -> np.ndarray:
    """Adjust tunable parameters for a given set of intermediate feature maps.

    Parameters
    ----------
    y : xr.DataArray
        Intermediate feature maps.
    threshold: float
        Spot detection threshold. Can be interpreted as the minimum number of fully confident pixels necessary to
        identify a spot. Default is 1.0.
    min_distance : int, optional
        Minimum distance between spots. Default is 1.

    Returns
    -------
    coords : np.ndarray
        Predicted spot coordinates.
    """

    if 'n' in y.dims:

        batch_axis_len = y.shape[0]
        coords = np.empty(batch_axis_len, dtype=object)
        for i in range(batch_axis_len):
            deltas = np.moveaxis(y[i, ..., :2, :, :].to_numpy(), -3, -1)
            pooled_labels = y[i, ..., 3, :, :].to_numpy()
            coords[i] = utils.compute_spot_coordinates(deltas, pooled_labels,
                                                       threshold=threshold, min_distance=min_distance)

    else:

        deltas = np.moveaxis(y[..., :2, :, :].to_numpy(), -3, -1)
        pooled_labels = y[..., 3, :, :].to_numpy()
        coords = utils.compute_spot_coordinates(deltas, pooled_labels, threshold=threshold, min_distance=min_distance)

    return coords



import jax
import jax.numpy as jnp

from functools import wraps
from jax import vmap
from typing import Callable, Optional, Union

from piscis.utils import smooth_sum_pool


def smoothf1_loss(
        deltas_pred: jax.Array,
        labels_pred: jax.Array,
        deltas: jax.Array,
        labels: jax.Array,
        dilation_iterations: int,
        max_distance: float,
        epsilon: float = 1e-7
) -> jax.Array:

    """Compute the SmoothF1 loss.

    Parameters
    ----------
    deltas_pred : jax.Array
        Predicted displacement vectors.
    labels_pred : jax.Array
        Predicted binary labels.
    deltas : jax.Array
        Ground truth displacement vectors.
    labels : jax.Array
        Ground truth binary labels.
    dilation_iterations : int
        Number of iterations used to dilate ground truth labels.
    max_distance : float
        Maximum distance for matching predicted and ground truth displacement vectors.
    epsilon : float, optional
        Small constant for numerical stability. Default is 1e-7.

    Returns
    -------
    smoothf1 : jax.Array
        SmoothF1 loss.
    """

    # Use labels as the support for deltas_pred.
    deltas_pred = labels * deltas_pred

    # Squeeze the channel dimension in labels arrays.
    labels_pred = labels_pred[:, :, 0]
    labels = labels[:, :, 0]

    # Apply deltas_pred to labels_pred.
    kernel_size = (2 * dilation_iterations + 1, ) * 2
    pooled_labels = smooth_sum_pool(deltas_pred, labels_pred, 0.5, kernel_size)
    distances = jnp.linalg.norm(deltas_pred - deltas, axis=-1)
    matches = jnp.maximum(1 - distances / max_distance, 0.0)

    # Estimate the number of true positives, false positives, and false negatives.
    tp = jnp.sum(pooled_labels * matches)
    fp = jnp.sum(pooled_labels) - tp
    fn = jnp.sum(labels) - tp

    # Compute the SmoothF1 loss.
    smoothf1 = -2 * tp / (2 * tp + fp + fn + epsilon)

    return smoothf1


def dice_loss(
        y_pred: jax.Array,
        y: jax.Array,
        epsilon: float = 1e-7
) -> jax.Array:

    """Compute the Dice loss.

    Parameters
    ----------
    y_pred : jax.Array
        Predicted binary labels.
    y : jax.Array
        Ground truth binary labels.
    epsilon : float, optional
        Small constant for numerical stability. Default is 1e-7.

    Returns
    -------
    dl : jax.Array
        Dice loss.
    """

    intersection = jnp.sum(y_pred * y)
    dl = - (2.0 * intersection) / (jnp.sum(y_pred) + jnp.sum(y) + epsilon)

    return dl


def masked_l2_loss(
        y_pred: jax.Array,
        y: jax.Array,
        mask: jax.Array
) -> jax.Array:

    """Compute the L2 loss over masked pixels.

    Parameters
    ----------
    y_pred : jax.Array
        Predicted values.
    y : jax.Array
        Ground truth values.
    mask : jax.Array
        Binary mask where 1 indicates pixels to compute the L2 loss and 0 indicates pixels to ignore.

    Returns
    -------
    rmse : jax.Array
        Masked root-mean-square error.
    """

    l2 = jnp.sum(jnp.linalg.norm(y_pred - y, axis=-1) * mask) / jnp.sum(mask)

    return l2


def mse_loss(
        y_pred: jax.Array,
        y: jax.Array,
) -> jax.Array:

    """Compute the mean squared error.

    Parameters
    ----------
    y_pred : jax.Array
        Predicted values.
    y : jax.Array
        Ground truth values.

    Returns
    -------
    mse : jax.Array
        Mean squared error.
    """

    mse = (y - y_pred) ** 2
    mse = reduce_loss(mse, 'mean')

    return mse


def bce_loss(
        y_pred: jax.Array,
        y: jax.Array,
        pos_weight: Union[float, jax.Array] = 1.0,
        epsilon: float = 1e-7,
        reduction: Optional[str] = 'mean'
) -> jax.Array:

    """Compute the binary cross entropy loss.

    Parameters
    ----------
    y_pred : jax.Array
        Predicted binary labels.
    y : jax.Array
        Ground truth binary labels.
    pos_weight : Union[float, jax.Array], optional
        Weight for positive class. Default is 1.0.
    epsilon : float, optional
        Small constant for numerical stability. Default is 1e-7.
    reduction : Optional[str], optional
        Loss reduction method. Default is 'mean'.

    Returns
    -------
    bce : jax.Array
        Binary cross entropy loss.
    """

    bce = -(pos_weight * jnp.log(y_pred + epsilon) * y + jnp.log((1 - y_pred) + epsilon) * (1 - y))
    bce = reduce_loss(bce, reduction)

    return bce


def bce_with_logits_loss(
        y_pred: jax.Array,
        y: jax.Array,
        pos_weight: Union[float, jax.Array] = 1.0,
        reduction: Optional[str] = 'mean'
) -> jax.Array:

    """Compute the binary cross entropy loss with logits.

    Parameters
    ----------
    y_pred : jax.Array
        Predicted logits.
    y : jax.Array
        Ground truth binary labels.
    pos_weight : Union[float, jax.Array], optional
        Weight for positive class. Default is 1.0.
    reduction : Optional[str], optional
        Loss reduction method. Default is 'mean'.

    Returns
    -------
    bce : jax.Array
        Binary cross entropy loss.
    """

    neg_abs = -jnp.abs(y_pred)
    bce = jnp.maximum(y_pred, 0) - y_pred * y + jnp.log(1 + jnp.exp(neg_abs))
    bce = jnp.where(y, pos_weight * bce, bce)
    bce = reduce_loss(bce, reduction)

    return bce


def weighted_bce_loss(
        y_pred: jax.Array,
        y: jax.Array,
        alpha: float = 1.0,
        epsilon: float = 1e-7,
        reduction: Optional[str] = 'mean'
) -> jax.Array:

    """Compute the weighted binary cross entropy loss.

    y_pred : jax.Array
        Predicted binary labels.
    y : jax.Array
        Ground truth binary labels.
    alpha : float, optional
        Exponent factor applied to the inverse class weight. Default is 1.0.
    epsilon : float, optional
        Small constant for numerical stability. Default is 1e-7.
    reduction : Optional[str], optional
        Loss reduction method. Default is 'mean'.

    Returns
    -------
    bce : jax.Array
        Weighted binary cross entropy loss.
    """

    pos_weight = _inverse_class_weight(y, alpha=alpha, epsilon=epsilon)
    bce = bce_loss(y_pred, y, pos_weight=pos_weight, epsilon=epsilon, reduction=reduction)

    return bce


def weighted_bce_with_logits_loss(
        y_pred: jax.Array,
        y: jax.Array,
        alpha: float = 1.0,
        epsilon: float = 1e-7,
        reduction: Optional[str] = 'mean'
) -> jax.Array:

    """Compute the weighted binary cross entropy loss with logits.

    y_pred : jax.Array
        Predicted logits.
    y : jax.Array
        Ground truth binary labels.
    alpha : float, optional
        Exponent factor applied to the inverse class weight. Default is 1.0.
    epsilon : float, optional
        Small constant for numerical stability. Default is 1e-7.
    reduction : Optional[str], optional
        Loss reduction method. Default is 'mean'.

    Returns
    -------
    bce : jax.Array
        Weighted binary cross entropy loss.
    """

    pos_weight = _inverse_class_weight(y, alpha=alpha, epsilon=epsilon)
    bce = bce_with_logits_loss(y_pred, y, pos_weight=pos_weight, reduction=reduction)

    return bce


def cb_bce_loss(
        y_pred: jax.Array,
        y: jax.Array,
        beta: float = 0.999,
        epsilon: float = 1e-7,
        reduction: Optional[str] = 'mean'
) -> jax.Array:

    """Compute the class-balanced binary cross entropy loss.

    Parameters
    ----------
    y_pred : jax.Array
        Predicted binary labels.
    y : jax.Array
        Ground truth binary labels.
    beta : float, optional
        Hyperparameter for computing the effective number of samples. Default is 0.999.
    epsilon : float, optional
        Small constant for numerical stability. Default is 1e-7.
    reduction : Optional[str], optional
        Loss reduction method. Default is 'mean'.

    Returns
    -------
    bce : jax.Array
        Class-balanced binary cross entropy loss.

    References
    ----------
    .. [1] Cui, Yin, et al. "Class-balanced loss based on effective number of samples." Proceedings of the IEEE/CVF
           conference on computer vision and pattern recognition. 2019.
    """

    pos_weight = _class_balanced_weight(y, beta=beta)
    bce = bce_loss(y_pred, y, pos_weight=pos_weight, epsilon=epsilon, reduction=reduction)

    return bce


def cb_bce_with_logit_loss(
        y_pred: jax.Array,
        y: jax.Array,
        beta: float = 0.999,
        reduction: Optional[str] = 'mean'
) -> jax.Array:

    """Compute the class-balanced binary cross entropy loss with logits.

    Parameters
    ----------
    y_pred : jax.Array
        Predicted logits.
    y : jax.Array
        Ground truth binary labels.
    beta : float, optional
        Hyperparameter for computing the effective number of samples. Default is 0.999.
    reduction : Optional[str], optional
        Loss reduction method. Default is 'mean'.

    Returns
    -------
    bce : jax.Array
        Class-balanced binary cross entropy loss.

    References
    ----------
    .. [1] Cui, Yin, et al. "Class-balanced loss based on effective number of samples." Proceedings of the IEEE/CVF
           conference on computer vision and pattern recognition. 2019.
    """

    pos_weight = _class_balanced_weight(y, beta=beta)
    bce = bce_with_logits_loss(y_pred, y, pos_weight=pos_weight, reduction=reduction)

    return bce


def la_loss(
        y_pred: jax.Array,
        y: jax.Array,
        tau: float = 1.0,
        reduction: Optional[str] = 'mean'
):

    """Compute the logit-adjusted loss.

    Parameters
    ----------
    y_pred : jax.Array
        Predicted logits.
    y : jax.Array
        Ground truth binary labels.
    tau : float, optional
        Hyperparameter for computing the logit adjustment. Default is 1.0.
    reduction : Optional[str], optional
        Loss reduction method. Default is 'mean'.

    Returns
    -------
    la : jax.Array
        Logit-adjusted loss.

    References
    ----------
    .. [1] Menon, Aditya Krishna, et al. "Long-tail learning via logit adjustment." arXiv preprint arXiv:2007.07314
           (2020).
    """

    y_pred = _logit_adjustment(y_pred, y, tau=tau)
    la = bce_with_logits_loss(y_pred, y, reduction=reduction)

    return la


def vs_loss(
        y_pred: jax.Array,
        y: jax.Array,
        tau: float = 1.0,
        gamma: float = 0.25,
        reduction: Optional[str] = 'mean'
):

    """Compute the vector-scaling loss.

    Parameters
    ----------
    y_pred : jax.Array
        Predicted logits.
    y : jax.Array
        Ground truth binary labels.
    tau : float, optional
        Hyperparameter for computing the additive logit adjustment. Default is 1.0.
    gamma : float, optional
        Hyperparameter for computing the multiplicative logit adjustment. Default is 0.25.
    reduction : Optional[str], optional
        Loss reduction method. Default is 'mean'.

    Returns
    -------
    vs : jax.Array
        Vector-scaling loss.

    References
    ----------
    .. [1] Kini, Ganesh Ramachandra, et al. "Label-imbalanced and group-sensitive classification under
           overparameterization." Advances in Neural Information Processing Systems 34 (2021): 18970-18983.
    """

    y_pred = _vector_scaling(y_pred, y, tau=tau, gamma=gamma)
    vs = bce_with_logits_loss(y_pred, y, reduction=reduction)

    return vs


def binary_focal_loss(
        y_pred: jax.Array,
        y: jax.Array,
        gamma: float = 2,
        pos_weight: Union[float, jax.Array] = 1.0,
        epsilon: float = 1e-7,
        reduction: Optional[str] = 'mean'
) -> jax.Array:

    """Compute the binary focal loss.

    Parameters
    ----------
    y_pred : jax.Array
        Predicted binary labels.
    y : jax.Array
        Ground truth binary labels.
    gamma : float, optional
        Hyperparameter for computing the binary focal loss. Default is 2.
    pos_weight : Union[float, jax.Array], optional
        Weight for positive class. Default is 1.0.
    epsilon : float, optional
        Small constant for numerical stability. Default is 1e-7.
    reduction : Optional[str], optional
        Loss reduction method. Default is 'mean'.

    Returns
    -------
    bf : jax.Array
        Binary focal loss.

    References
    ----------
    .. [1] Lin, Tsung-Yi, et al. "Focal loss for dense object detection." Proceedings of the IEEE international
           conference on computer vision. 2017.
    """

    bce = bce_loss(y_pred, y, pos_weight=pos_weight, epsilon=epsilon, reduction=reduction)
    p_t = y_pred * y + (1 - y_pred) * (1 - y)
    bf = bce * (1 - p_t) ** gamma
    bf = reduce_loss(bf, reduction=reduction)

    return bf


def weighted_binary_focal_loss(
        y_pred: jax.Array,
        y: jax.Array,
        alpha: float = 1.0,
        epsilon: float = 1e-7,
        reduction: Optional[str] = 'mean'
) -> jax.Array:

    """Compute the weighted binary focal loss.

    y_pred : jax.Array
        Predicted binary labels.
    y : jax.Array
        Ground truth binary labels.
    alpha : float, optional
        Exponent factor applied to the inverse class weight. Default is 1.0.
    epsilon : float, optional
        Small constant for numerical stability. Default is 1e-7.
    reduction : Optional[str], optional
        Loss reduction method. Default is 'mean'.

    Returns
    -------
    bce : jax.Array
        Weighted binary focal loss.
    """

    pos_weight = _inverse_class_weight(y, alpha=alpha, epsilon=epsilon)
    bf = binary_focal_loss(y_pred, y, pos_weight=pos_weight, epsilon=epsilon, reduction=reduction)

    return bf


def ce_loss(
        y_pred: jax.Array,
        y: jax.Array,
        epsilon: float = 1e-7,
        reduction: Optional[str] = 'mean'
) -> jax.Array:

    """Compute the cross entropy loss.

    Parameters
    ----------
    y_pred : jax.Array
        Predicted labels.
    y : jax.Array
        Ground truth labels.
    epsilon : float, optional
        Small constant for numerical stability. Default is 1e-7.
    reduction : Optional[str], optional
        Loss reduction method. Default is 'mean'.

    Returns
    -------
    ce : jax.Array
        Cross entropy loss.
    """

    ce = -(jnp.log(y_pred + epsilon) * y)
    ce = reduce_loss(ce, reduction=reduction)

    return ce


def focal_loss(
        y_pred: jax.Array,
        y: jax.Array,
        gamma: float = 2,
        epsilon: float = 1e-7,
        reduction: Optional[str] = 'mean'
) -> jax.Array:

    """Compute the focal loss.

    Parameters
    ----------
    y_pred : jax.Array
        Predicted labels.
    y : jax.Array
        Ground truth labels.
    gamma : float, optional
        Hyperparameter for computing the focal loss. Default is 2.
    epsilon : float, optional
        Small constant for numerical stability. Default is 1e-7.
    reduction : Optional[str], optional
        Loss reduction method. Default is 'mean'.

    Returns
    -------
    f : jax.Array
        Focal loss.

    References
    ----------
    .. [1] Lin, Tsung-Yi, et al. "Focal loss for dense object detection." Proceedings of the IEEE international
           conference on computer vision. 2017.
    """

    f = -((1 - y_pred) ** gamma * jnp.log(y_pred + epsilon) * y)
    f = reduce_loss(f, reduction=reduction)

    return f


def _inverse_class_weight(
        y: jax.Array,
        alpha: float,
        epsilon: float
) -> jax.Array:

    """Compute the inverse class weight.

    Parameters
    ----------
    y : jax.Array
        Ground truth binary labels.
    alpha : float
        Exponent of the inverse class weight.
    epsilon : float
        Small constant for numerical stability.

    Returns
    -------
    pos_weight : jax.Array
        Inverse class weight.
    """

    pos_weight = (jnp.sum(~y) / (jnp.sum(y) + epsilon)) ** alpha

    return pos_weight


def _class_balanced_weight(
        y: jax.Array,
        beta: float
) -> jax.Array:

    """Compute the class-balanced weight.

    Parameters
    ----------
    y : jax.Array
        Ground truth binary labels.
    beta : float
        Hyperparameter for computing the effective number of samples.

    Returns
    -------
    pos_weight : jax.Array
        Class-balanced weight.

    References
    ----------
    .. [1] Cui, Yin, et al. "Class-balanced loss based on effective number of samples." Proceedings of the IEEE/CVF
           conference on computer vision and pattern recognition. 2019.
    """

    pos_weight = (1 - beta ** jnp.sum(~y)) / (1 - beta ** jnp.sum(y))

    return pos_weight


def _logit_adjustment(
        y_pred: jax.Array,
        y: jax.Array,
        tau: float
) -> jax.Array:

    """Apply logit adjustment to the predicted logits.

    Parameters
    ----------
    y_pred : jax.Array
        Predicted logits.
    y : jax.Array
        Ground truth binary labels.
    tau : float
        Hyperparameter for computing the logit adjustment.

    Returns
    -------
    y_pred : jax.Array
        Adjusted logits.

    References
    ----------
    .. [1] Menon, Aditya Krishna, et al. "Long-tail learning via logit adjustment." arXiv preprint arXiv:2007.07314
           (2020).
    """

    # Compute class sizes.
    num_pos = jnp.sum(y)
    num_neg = jnp.sum(~y)
    num_tot = num_pos + num_neg

    # Compute logit adjustments.
    iota_pos = tau * jnp.log(num_pos / num_tot)
    iota_neg = tau * jnp.log(num_neg / num_tot)

    # Apply logit adjustments.
    y_pred = jnp.where(y, y_pred + iota_pos, y_pred + iota_neg)

    return y_pred


def _vector_scaling(
        y_pred: jax.Array,
        y: jax.Array,
        tau: float,
        gamma: float
) -> jax.Array:

    """Apply vector scaling to the predicted logits.

    Parameters
    ----------
    y_pred : jax.Array
        Predicted logits.
    y : jax.Array
        Ground truth binary labels.
    tau : float
        Hyperparameter for computing the additive logit adjustment.
    gamma : float
        Hyperparameter for computing the multiplicative logit adjustment.

    Returns
    -------
    y_pred : jax.Array
        Adjusted logits.

    References
    ----------
    .. [1] Kini, Ganesh Ramachandra, et al. "Label-imbalanced and group-sensitive classification under
           overparameterization." Advances in Neural Information Processing Systems 34 (2021): 18970-18983.
    """

    # Compute class sizes.
    num_pos = jnp.sum(y)
    num_neg = jnp.sum(~y)
    num_tot = num_pos + num_neg
    num_max = jnp.maximum(num_pos, num_neg)

    # Compute additive logit adjustments.
    iota_pos = tau * jnp.log(num_pos / num_tot)
    iota_neg = tau * jnp.log(num_neg / num_tot)

    # Compute multiplicative logit adjustments.
    delta_pos = (num_pos / num_max) ** gamma
    delta_neg = (num_neg / num_max) ** gamma

    # Apply vector scaling.
    y_pred = jnp.where(y, delta_pos * y_pred + iota_pos, delta_neg * y_pred + iota_neg)

    return y_pred


def wrap_loss_fn(
        loss_fn: Callable,
        axis: int = 0,
        reduction: Optional[str] = 'mean'
) -> Callable:
    """Wrap a loss function for vectorization and loss reduction.

    Parameters
    ----------
    loss_fn : Callable
        Loss function.
    axis : int, optional
        Axis to vectorize over. Default is 0.
    reduction : Optional[str], optional
        Loss reduction method. Default is 'mean'.

    Returns
    -------
    wrapped_loss_fn : Callable
        Wrapped loss function.
    """

    @wraps(loss_fn)
    def wrapped_loss_fn(*args):

        # Build in_axes dynamically from the arguments.
        in_axes = tuple(axis if isinstance(arg, jax.Array) else None for arg in args)

        # Vectorize the loss function.
        loss = vmap(loss_fn, in_axes=in_axes)(*args)

        # Reduce the loss.
        loss = reduce_loss(loss, reduction=reduction)

        return loss

    return wrapped_loss_fn


def reduce_loss(
        loss: jax.Array,
        reduction: Optional[str] = 'mean'
) -> jax.Array:

    """Reduce the loss.

    Parameters
    ----------
    loss : jax.Array
        Loss array to be reduced.
    reduction : Optional[str], optional
        Loss reduction method. Supported methods are 'mean' and 'sum'. Default is 'mean'.

    Returns
    -------
    loss : jax.Array
        Reduced loss.

    Raises
    ------
    ValueError
        If the `reduction` method is not supported.
    """

    if reduction == 'mean':
        loss = jnp.mean(loss)
    elif reduction == 'sum':
        loss = jnp.sum(loss)
    else:
        raise ValueError("Reduction method is not supported.")

    return loss



import optax

from typing import Any, Callable, Optional, Union


def sgdw(
    learning_rate: optax.ScalarOrSchedule,
    momentum: Optional[float] = None,
    nesterov: bool = False,
    accumulator_dtype: Optional[Any] = None,
    weight_decay: float = 1e-4,
    mask: Optional[Union[Any, Callable[[optax.Params], Any]]] = None,
) -> optax.GradientTransformation:

    """A canonical Stochastic Gradient Descent optimizer.

    This implements stochastic gradient descent. It also includes support for momentum, and nesterov acceleration, as
    these are standard practice when using stochastic gradient descent to train deep neural networks.

    Parameters
    ----------
    learning_rate : Union[optax.Schedule, float]
        A fixed global scaling factor.
    momentum : Optional[float], optional
        Decay rate used by the momentum term. If None, momentum is not used at all. Default is None.
    nesterov : bool, optional
        Whether Nesterov momentum is used. Default is False.
    accumulator_dtype : Optional[Any], optional
        Optional `dtype` to be used for the accumulator. If None, the `dtype` is inferred from `params` and `updates`.
        Default is None.
    weight_decay : float, optional
        Strength of the weight decay regularization. Note that this weight decay is multiplied with the learning rate.
        This is consistent with other frameworks such as PyTorch, but different from (Loshchilov et al., 2019) where the
        weight decay is only multiplied with the "schedule multiplier", but not the base learning rate.
    mask : Optional[Union[Any, Callable[[optax.Params], Any]]], optional
        A tree with same structure as (or a prefix of) the params PyTree, or a Callable that returns such a pytree given
        the params/updates. The leaves should be booleans, True for leaves/subtrees you want to apply the weight decay
        to, and False for those you want to skip. Note that the SGDW gradient transformations are applied to all
        parameters.

    Returns
    -------
    tx : optax.GradientTransformation
        SGDW gradient transformation.

    References
    ----------
    .. [1] Sutskever, Ilya, et al. "On the importance of initialization and momentum in deep learning." International
           conference on machine learning. PMLR, 2013.
    .. [2] Ilya, Loshchilov, and Hutter Frank. "Decoupled weight decay regularization." Proceedings of ICLR 7 (2019).
    """

    tx = optax.chain(
        (
            optax.trace(
                decay=momentum, nesterov=nesterov, accumulator_dtype=accumulator_dtype
            )
            if momentum is not None
            else optax.identity()
        ),
        optax.add_decayed_weights(weight_decay, mask),
        optax.scale_by_learning_rate(learning_rate),
    )

    return tx



import jax
import jax.numpy as jnp

from flax import linen as nn
from functools import partial
from typing import Any, Callable, Optional, Sequence

ModuleDef = Any


class Conv(nn.Module):

    """Convolutional block with batch norm and activation.

    Attributes
    ----------
    features : int
        Number of output features.
    kernel_size : Sequence[int]
        Size of the convolutional kernel.
    conv : Optional[ModuleDef]
        Convolution module.
    bn : Optional[ModuleDef]
        Batch norm module.
    act : Optional[Callable]
        Activation function.
    layers : Sequence[str]
        Sequence of layers to apply.
    """

    features: int
    kernel_size: Sequence[int]
    conv: Optional[ModuleDef]
    bn: Optional[ModuleDef]
    act: Optional[Callable]
    layers: Sequence[str]

    @nn.compact
    def __call__(self, x: jax.Array) -> jax.Array:

        conv = partial(
            self.conv,
            features=self.features,
            kernel_size=self.kernel_size,
        )

        layers = []

        for layer in self.layers:
            if layer == 'conv':
                layers.append(conv())
            elif layer == 'bn':
                layers.append(self.bn())
            elif layer == 'act':
                layers.append(self.act)
            else:
                raise ValueError(f"Unknown layer {layer}.")

        x = nn.Sequential(layers)(x)

        return x


class MBConv(nn.Module):

    """Mobile inverted bottleneck convolutional block.

    Attributes
    ----------
    features_in : int
        Number of input features.
    features_out : int
        Number of output features.
    expand_ratio : int
        Expansion ratio.
    kernel_size : Sequence[int]
        Size of the convolutional kernel.
    strides : int
        Stride of the convolution.
    se_ratio : float
        Squeeze and excitation ratio.
    conv : ModuleDef
        Convolution module.
    dropout : Optional[ModuleDef]
        Dropout module.
    bn : ModuleDef
        Batch norm module.
    act : Callable
        Activation function.
    """

    features_in: int
    features_out: int
    expand_ratio: int
    kernel_size: Sequence[int]
    strides: int
    se_ratio: float
    conv: ModuleDef
    dropout: Optional[ModuleDef]
    bn: ModuleDef
    act: Callable

    @nn.compact
    def __call__(self, x: jax.Array) -> jax.Array:

        conv = partial(
            self.conv,
            kernel_size=self.kernel_size,
            strides=self.strides,
            padding='SAME'
        )

        proj = partial(
            self.conv,
            kernel_size=(1, 1),
            strides=1,
            padding='SAME'
        )

        inputs = x

        # Expansion phase.
        filters = self.features_in * self.expand_ratio
        if self.expand_ratio != 1:

            x = proj(
                features=filters,
                use_bias=False,
                name='expand_conv',
            )(x)
            x = self.bn(
                name='expand_bn'
            )(x)
            x = self.act(x)

        # Depthwise convolution.
        x = conv(
            features=x.shape[-1],
            feature_group_count=x.shape[-1],
            use_bias=False,
            name='dw_conv'
        )(x)
        x = self.bn(
            name='dw_bn'
        )(x)
        x = self.act(x)

        # Squeeze and excite.
        if 0 < self.se_ratio <= 1:

            filters_se = max(1, int(self.features_in * self.se_ratio))
            se = jnp.mean(x, axis=(1, 2), keepdims=True)

            se = proj(
                features=filters_se,
                use_bias=True,
                name='se_reduce'
            )(se)
            se = self.act(se)
            se = proj(
                features=filters,
                use_bias=True,
                name='se_expand'
            )(se)
            se = nn.sigmoid(se)

            x = x * se

        # Output phase.
        x = proj(
            features=self.features_out,
            use_bias=False,
            name='project_conv'
        )(x)
        x = self.bn(
            name='project_bn'
        )(x)

        # Residual.
        if (self.strides == 1) and (self.features_in == self.features_out):

            if self.dropout is not None:
                x = self.dropout(
                    name='drop'
                )(x)
            x = x + inputs

        return x


class FusedMBConv(nn.Module):

    """Fused mobile inverted bottleneck convolutional block.

    Attributes
    ----------
    features_in : int
        Number of input features.
    features_out : int
        Number of output features.
    expand_ratio : int
        Expansion ratio.
    kernel_size : Sequence[int]
        Size of the convolutional kernel.
    strides : int
        Stride of the convolution.
    se_ratio : float
        Squeeze and excitation ratio.
    conv : ModuleDef
        Convolution module.
    dropout : Optional[ModuleDef]
        Dropout module.
    bn : ModuleDef
        Batch norm module.
    act : Callable
        Activation function.
    """

    features_in: int
    features_out: int
    expand_ratio: int
    kernel_size: Sequence[int]
    strides: int
    se_ratio: float
    conv: ModuleDef
    dropout: Optional[ModuleDef]
    bn: ModuleDef
    act: Callable

    @nn.compact
    def __call__(self, x: jax.Array) -> jax.Array:

        conv = partial(
            self.conv,
            kernel_size=self.kernel_size,
            strides=self.strides,
            padding='SAME'
        )

        proj = partial(
            self.conv,
            kernel_size=(1, 1),
            strides=1,
            padding='SAME'
        )

        inputs = x

        # Expansion phase.
        filters = self.features_in * self.expand_ratio
        if self.expand_ratio != 1:

            x = conv(
                features=filters,
                use_bias=False,
                name='expand_conv',
            )(x)
            x = self.bn(
                name='expand_bn'
            )(x)
            x = self.act(x)

        # Squeeze and excite.
        if 0 < self.se_ratio <= 1:

            filters_se = max(1, int(self.features_in * self.se_ratio))
            se = jnp.mean(x, axis=(1, 2), keepdims=True)

            se = proj(
                features=filters_se,
                use_bias=True,
                name='se_reduce'
            )(se)
            se = self.act(se)
            se = proj(
                features=filters,
                use_bias=True,
                name='se_expand'
            )(se)
            se = nn.sigmoid(se)

            x = x * se

        # Output phase.
        if self.expand_ratio == 1:
            x = conv(
                features=self.features_out,
                use_bias=False,
                name='project_conv'
            )(x)
        else:
            x = proj(
                features=self.features_out,
                use_bias=False,
                name='project_conv'
            )(x)
        x = self.bn(
            name='project_bn'
        )(x)
        if self.expand_ratio == 1:
            x = self.act(x)

        # Residual.
        if (self.strides == 1) and (self.features_in == self.features_out):

            if self.dropout is not None:
                x = self.dropout(
                    name='drop'
                )(x)
            x = x + inputs

        return x



import jax
import jax.numpy as jnp

from flax import linen as nn
from functools import partial
from jax.image import resize
from typing import Any, Callable, Optional, Sequence, Tuple

from piscis.networks.conv import Conv

ModuleDef = Any

BatchActConv = partial(Conv, layers=['bn', 'act', 'conv'])
BatchConv = partial(Conv, layers=['bn', 'conv'], act=None)


class BatchConvStyle(nn.Module):

    """Convolutional block with batch norm, activation, and style transfer.

    Attributes
    ----------
    features : int
        Number of output features.
    kernel_size : Sequence[int]
        Size of the convolutional kernel.
    conv : ModuleDef
        Convolution module.
    dense : ModuleDef
        Dense module.
    dropout : ModuleDef
        Dropout module.
    bn : ModuleDef
        Batch norm module.
    act : Callable
        Activation function.
    """

    features: int
    kernel_size: Sequence[int]
    conv: ModuleDef
    dense: ModuleDef
    dropout: ModuleDef
    bn: ModuleDef
    act: Callable

    @nn.compact
    def __call__(
            self,
            style: Optional[jax.Array],
            x: jax.Array,
            y: Optional[jax.Array] = None
    ) -> jax.Array:

        conv = partial(
            BatchActConv,
            features=self.features,
            kernel_size=self.kernel_size,
            conv=self.conv,
            bn=self.bn,
            act=self.act
        )

        if y is not None:
            x = self.dropout()(x)
            x = x + y

        if style is not None:
            full = partial(
                self.dense,
                features=self.features
            )
            feat = full()(style)
            x = x + feat[:, None, None]

        x = conv()(x)

        return x


class UpConv(nn.Module):

    """Upsampling convolutional block.

    Attributes
    ----------
    features : int
        Number of output features.
    kernel_size : Sequence[int]
        Size of the convolutional kernel.
    conv : ModuleDef
        Convolution module.
    dense : ModuleDef
        Dense module.
    dropout : ModuleDef
        Dropout module.
    bn : ModuleDef
        Batch norm module.
    act : Callable
        Activation function.
    """

    features: int
    kernel_size: Sequence[int]
    conv: ModuleDef
    dense: ModuleDef
    dropout: ModuleDef
    bn: ModuleDef
    act: Callable

    @nn.compact
    def __call__(
            self,
            x: jax.Array,
            y: Optional[jax.Array],
            style: Optional[jax.Array]
    ) -> jax.Array:

        conv = partial(
            BatchActConv,
            features=self.features,
            kernel_size=self.kernel_size,
            conv=self.conv,
            bn=self.bn,
            act=self.act
        )

        convs = partial(
            BatchConvStyle,
            features=self.features,
            kernel_size=self.kernel_size,
            conv=self.conv,
            dense=self.dense,
            dropout=self.dropout,
            bn=self.bn,
            act=self.act
        )

        proj = partial(
            Conv,
            features=self.features,
            kernel_size=(1, 1),
            conv=self.conv,
            bn=self.bn,
            act=None,
            layers=['bn', 'conv']
        )

        x = proj()(x) + convs()(style, conv()(x), y)
        x = x + convs()(style, convs()(style, x))

        return x


class MakeStyle(nn.Module):

    """Style transfer module."""

    @nn.compact
    def __call__(self, x: jax.Array) -> jax.Array:

        style = jnp.mean(x, axis=(1, 2))
        style = style / jnp.linalg.norm(style, axis=1, keepdims=True)

        return style


class Decoder(nn.Module):

    """Decoder module.

    Attributes
    ----------
    kernel_size : Sequence[int]
        Size of the convolutional kernel.
    aggregate : str
        Aggregation mode for feature maps. Supported modes are 'sum' and 'concatenate'.
    conv : ModuleDef
        Convolution module.
    dense : ModuleDef
        Dense module.
    dropout : ModuleDef
        Dropout module.
    bn : ModuleDef
        Batch norm module.
    act : Callable
        Activation function.
    """

    kernel_size: Sequence[int]
    aggregate: str
    conv: ModuleDef
    dense: ModuleDef
    dropout: ModuleDef
    bn: ModuleDef
    act: Callable

    @nn.compact
    def __call__(
            self,
            style: jax.Array,
            xd: Sequence[jax.Array]
    ) -> jax.Array:

        up = partial(
            UpConv,
            kernel_size=self.kernel_size,
            conv=self.conv,
            dense=self.dense,
            dropout=self.dropout,
            bn=self.bn,
            act=self.act
        )

        stage_sizes = [x.shape[-1] for x in xd]

        # Create a list to store feature maps.
        feature_maps = []

        f = up(
            features=stage_sizes[-1],
        )(xd[-1], xd[-1], style)
        feature_maps.append(f)

        for i, features in reversed(list(enumerate(stage_sizes[:-1]))):
            f = _interpolate(f, scale=2, method='nearest')
            f = up(
                features=features,
            )(f, xd[i], style)
            feature_maps.append(f)

        # Resize feature maps.
        for i in range(len(feature_maps[:-1])):
            for j in range(len(feature_maps) - i - 1):
                feature_maps[i] = _interpolate(feature_maps[i], scale=2, method='nearest')
                feature_maps[i] = up(
                    features=stage_sizes[0]
                )(feature_maps[i], None, style)

        # Aggregate feature maps.
        if self.aggregate == 'sum':
            aggregate_feature_maps = jnp.sum(jnp.array(feature_maps), axis=0)
        elif self.aggregate == 'concatenate':
            aggregate_feature_maps = jnp.concatenate(feature_maps, axis=-1)
        else:
            raise ValueError(f"Aggregation mode is not supported.")

        return aggregate_feature_maps


class FPN(nn.Module):

    """Feature pyramid network.

    Attributes
    ----------
    encoder : ModuleDef
        Encoder module.
    encoder_levels : Sequence[int]
        Encoder levels to use for the feature pyramid.
    features : int
        Number of output features.
    kernel_size : Sequence[int]
        Size of the convolutional kernel.
    style : bool
        Whether to use style transfer.
    aggregate : str
        Aggregation mode for feature maps. Supported modes are 'sum' and 'concatenate'.
    dropout_rate : float
        Dropout rate at skip connections.
    bn_momentum : float
        Momentum parameter for batch norm layers.
    bn_epsilon : float
        Epsilon parameter for batch norm layers.
    conv : ModuleDef
        Convolution module.
    dense : ModuleDef
        Dense module.
    bn : ModuleDef
        Batch norm module.
    act : Callable
        Activation function.
    """

    encoder: ModuleDef
    encoder_levels: Sequence[int]
    features: int
    kernel_size: Sequence[int] = (3, 3)
    style: bool = True
    aggregate: str = 'sum'
    dropout_rate: float = 0.2
    bn_momentum: float = 0.9
    bn_epsilon: float = 1e-05
    conv: ModuleDef = nn.Conv
    dense: ModuleDef = nn.Dense
    dropout: ModuleDef = nn.Dropout
    bn: ModuleDef = nn.BatchNorm
    act: Callable = nn.swish

    @nn.compact
    def __call__(
            self,
            x: jax.Array,
            train: bool = True
    ) -> Tuple[jax.Array, Optional[jax.Array]]:

        dropout = partial(
            self.dropout,
            rate=self.dropout_rate,
            deterministic=not train
        )

        bn = partial(
            self.bn,
            use_running_average=not train,
            momentum=self.bn_momentum,
            epsilon=self.bn_epsilon,
        )

        decoder = partial(
            Decoder,
            kernel_size=self.kernel_size,
            aggregate=self.aggregate,
            conv=self.conv,
            dense=self.dense,
            dropout=dropout,
            bn=bn,
            act=self.act
        )

        output = partial(
            BatchActConv,
            features=self.features,
            kernel_size=(1, 1),
            conv=self.conv,
            bn=bn,
            act=self.act
        )

        # Get encoder outputs.
        x = self.encoder()(x, train=train, capture_list=self.encoder_levels)
        x = [x[i] for i in self.encoder_levels]

        # Make style vectors if necessary.
        if self.style:
            style = MakeStyle()(x[-1])
        else:
            style = None

        # Apply decoder and output block.
        x = decoder()(style, x)
        x = output()(x)

        return x, style


def _interpolate(
        x: jax.Array,
        scale: float,
        *args: Any,
        **kwargs: Any
):

    """Interpolate feature maps.

    Parameters
    ----------
    x : jax.Array
        Feature maps.
    scale : float
        Scale factor.
    *args : Any
        Positional arguments for the resize function.
    **kwargs : Any
        Keyword arguments for the resize function.

    Returns
    -------
    x : jax.Array
        Interpolated feature maps.
    """

    x = resize(x, (x.shape[0], round(x.shape[1] * scale), round(x.shape[2] * scale), x.shape[3]), *args, **kwargs)

    return x



import cv2 as cv
import jax
import jax.numpy as jnp
import numpy as np

from jax import jit, random, vmap
from scipy import ndimage
from typing import Any, List, Optional, Sequence, Tuple


class RandomAugment:

    """Transformer for random data augmentation.

    Attributes
    ----------
    output_size : Optional[Tuple[int, int]]
        Output size of the transformer.
    scales : List[float]
        List of image scales.
    dxys : List[Tuple[float, float]]
        List of image translation vectors.
    thetas : List[float]
        List of image rotation angles.
    affines : List[np.ndarray]
        List of affine transformation matrices.
    flips0 : List[bool]
        List of booleans for flipping along axis 0.
    flips1 : List[bool]
        List of booleans for flipping along axis 1.
    intensity_scales : List[float]
        List of image intensity scale factors.
    """

    def __init__(self) -> None:

        """Initialize a RandomAugment object."""

        self.output_size = None
        self.scales = []
        self.dxys = []
        self.thetas = []
        self.affines = []
        self.flips0 = []
        self.flips1 = []
        self.intensity_scales = []

    def generate_transforms(
            self,
            images: Sequence[np.ndarray],
            key: jax.Array,
            base_scales: Sequence[float],
            output_size: Tuple[int, int],
            min_scale_factor: float = 0.75,
            max_scale_factor: float = 1.25,
            max_intensity_scale_factor: float = 5
    ) -> None:

        """Generate random transformations.

        Parameters
        ----------
        images : Sequence[np.ndarray]
            Images to transform.
        key : jax.Array
            Random key used for generating random transformations.
        base_scales : Sequence[float]
            List of base scales for each image.
        output_size : Tuple[int, int]
            Output size of the transformer.
        min_scale_factor : float, optional
            Minimum scale factor. Default is 0.75.
        max_scale_factor : float, optional
            Maximum scale factor. Default is 1.25.
        max_intensity_scale_factor : float, optional
            Maximum intensity scale factor. Intensity scale factors are sampled from a log-uniform distribution with
            support on the interval [1 / `max_intensity_scale_factor`, `max_intensity_scale_factor`]. Default is 5.
        """

        # Reset class attributes.
        self.output_size = output_size
        self.flips0 = []
        self.flips1 = []
        self.scales = []
        self.dxys = []
        self.thetas = []
        self.affines = []
        self.intensity_scales = []

        # Determine the number of channels.
        channels = images[0].shape[-1]

        for image, base_scale in zip(images, base_scales):

            # Random flips.
            key, *subkeys = random.split(key, 3)
            flip0 = random.uniform(subkeys[0]) > 0.5
            self.flips0.append(float(flip0))
            flip1 = random.uniform(subkeys[1]) > 0.5
            self.flips1.append(float(flip1))

            # Random scaling.
            key, subkey = random.split(key)
            scale = base_scale * (min_scale_factor + (max_scale_factor - min_scale_factor) * random.uniform(subkey))
            self.scales.append(float(scale))

            # Random translation.
            key, subkey = random.split(key)
            dxy = np.maximum(0, np.array([image.shape[1] * scale - output_size[1],
                                          image.shape[0] * scale - output_size[0]]))
            dxy = (random.uniform(subkey, (2, )) - 0.5) * dxy
            self.dxys.append(np.asarray(dxy))

            # Random rotation.
            key, subkey = random.split(key)
            theta = random.uniform(subkey) * 2 * np.pi
            self.thetas.append(float(theta))

            # Construct affine transformation.
            image_center = (image.shape[1] / 2, image.shape[0] / 2)
            affine = cv.getRotationMatrix2D(image_center, float(theta * 180 / np.pi), float(scale))
            affine[:, 2] += np.array(output_size) / 2 - np.array(image_center) + dxy
            self.affines.append(affine)

            # Random intensity scaling.
            key, subkey = random.split(key)
            intensity_scale = np.exp(
                (random.uniform(subkey, shape=(1, 1, channels)) - 0.5) * 2 * np.log(max_intensity_scale_factor)
            )
            self.intensity_scales.append(intensity_scale)

    def apply_coord_transforms(
            self,
            coords: Sequence[np.ndarray],
            filter_coords: bool = True
    ) -> List[np.ndarray]:

        """Apply random transformations to coordinates.

        Parameters
        ----------
        coords : Sequence[np.ndarray]
            Coordinates to transform.
        filter_coords : bool, optional
            Whether to filter coordinates outside the transformed image. Default is True.

        Returns
        -------
        transformed_coords : List[np.ndarray]
            Transformed coordinates.
        """

        transformed_coords = []

        for coord, flip0, flip1, affine in zip(coords, self.flips0, self.flips1, self.affines):

            # Apply affine transformation
            coord = np.concatenate((np.flip(coord, axis=1), np.ones((len(coord), 1))), axis=1)
            transformed_coord = np.flip((coord @ affine.T), axis=1)

            # Random flip
            if flip0:
                transformed_coord[:, 0] = self.output_size[0] - 1 - transformed_coord[:, 0]
            if flip1:
                transformed_coord[:, 1] = self.output_size[1] - 1 - transformed_coord[:, 1]

            # Filter coordinates outside transformed image
            if filter_coords:
                transformed_coord = transformed_coord[np.all((transformed_coord > -0.5) &
                                                             (transformed_coord < np.array(self.output_size) - 0.5),
                                                             axis=1)]

            transformed_coords.append(transformed_coord)

        return transformed_coords

    def apply_image_transforms(
            self,
            images: Sequence[np.ndarray],
            interpolation: str = 'nearest'
    ) -> List[np.ndarray]:

        """Apply random transformations to images.

        Parameters
        ----------
        images : Sequence[np.ndarray]
            Images to transform.
        interpolation : str, optional
            Interpolation mode for image scaling. Supported modes are 'nearest' or 'bilinear'. Default is 'nearest'.

        Returns
        -------
        transformed_images : List[np.ndarray]
            Transformed images.

        Raises
        ------
        ValueError
            If the `interpolation` mode is not supported.
        """

        transformed_images = []

        for image, affine, flip0, flip1, intensity_scale in \
                zip(images, self.affines, self.flips0, self.flips1, self.intensity_scales):

            # Apply affine transformation
            if interpolation == 'nearest':
                image = cv.warpAffine(image, M=affine, dsize=self.output_size, flags=cv.INTER_NEAREST)
            elif interpolation == 'bilinear':
                image = cv.warpAffine(image, M=affine, dsize=self.output_size, flags=cv.INTER_LINEAR)
            else:
                raise ValueError("Interpolation mode is not supported.")

            # Add channel axis if necessary.
            if image.ndim == 2:
                image = image[:, :, None]

            # Random flip
            if flip0:
                image = np.flip(image, axis=0)
            if flip1:
                image = np.flip(image, axis=1)

            # Random intensity scaling
            image = image * intensity_scale

            transformed_images.append(image)

        return transformed_images


def batch_adjust(
        images: Sequence[np.ndarray],
        adjustment: Optional[str],
        **kwargs: Any
) -> Sequence[np.ndarray]:

    """Batch adjust images.

    Parameters
    ----------
    images : Sequence[np.ndarray]
        Images to adjust.
    adjustment : Optional[str]
        Adjustment type. Supported types are 'normalize' and 'standardize'.
    **kwargs : Any
        Keyword arguments for the adjustment function.

    Returns
    -------
    adjusted_images : Sequence[np.ndarray]
        Adjusted images.
    """

    if adjustment is not None:
        images = list(images)
        adjusted_images = np.empty(len(images), dtype=object)
        for i, image in enumerate(images):
            adjusted_images[i] = adjust(image, adjustment, **kwargs)
    else:
        adjusted_images = images

    return adjusted_images


def adjust(
        image: np.ndarray,
        adjustment: Optional[str],
        **kwargs: Any
) -> np.ndarray:

    """Adjust an image.

    Parameters
    ----------
    image : np.ndarray
        Image to adjust.
    adjustment : Optional[str]
        Adjustment type. Supported types are 'normalize' and 'standardize'.
    **kwargs : Any
        Keyword arguments for the adjustment function.

    Returns
    -------
    adjusted_image : np.ndarray
        Adjusted image.

    Raises
    ------
    ValueError
        If the `adjustment` type is not supported.
    """

    if adjustment is None:
        adjusted_image = image
    elif adjustment == 'normalize':
        adjusted_image = normalize(image, **kwargs)
    elif adjustment == 'standardize':
        adjusted_image = standardize(image, **kwargs)
    else:
        raise ValueError("Adjustment type is not supported.")

    return adjusted_image


def normalize(
        image: np.ndarray,
        lower: float = 0,
        upper: float = 100,
        epsilon: float = 1e-7
) -> np.ndarray:

    """Normalize an image to the range [0, 1] based on the specified percentiles.

    Parameters
    ----------
    image : np.ndarray
        Image to normalize.
    lower : float, optional
        Lower percentile. Default is 0.
    upper : float, optional
        Upper percentile. Default is 100.
    epsilon : float, optional
        Small constant for numerical stability. Default is 1e-7.

    Returns
    -------
    normalized_image : np.ndarray
        Normalized image.
    """

    image_lower = np.percentile(image, lower, axis=(-2, -1), keepdims=True)
    image_upper = np.percentile(image, upper, axis=(-2, -1), keepdims=True)
    normalized_image = (image - image_lower) / (image_upper - image_lower + epsilon)

    return normalized_image


def standardize(
        image: np.ndarray,
        epsilon: float = 1e-7
) -> np.ndarray:

    """Standardize an image to zero mean and unit variance.

    Parameters
    ----------
    image : np.ndarray
        Image to standardize.
    epsilon : float, optional
        Small constant for numerical stability. Default is 1e-7.

    Returns
    -------
    standardized_image : np.ndarray
        Standardized image.
    """

    image_mean = image.mean(axis=(-2, -1), keepdims=True)
    image_std = image.std(axis=(-2, -1), keepdims=True)
    standardized_image = (image - image_mean) / (image_std + epsilon)

    return standardized_image


def voronoi_transform(
        coords: Sequence[np.ndarray],
        output_size: Tuple[int, int] = (256, 256),
        dilation_iterations: int = 1,
        coords_pad_length: Optional[int] = None
) -> Tuple[np.ndarray, np.ndarray]:

    """Transform a list of coordinates to generate ground truth binary labels and displacement vectors from each pixel
    to the nearest point via a Voronoi tessellation. Adapted from DeepCell Spots.

    Parameters
    ----------
    coords : Sequence[np.ndarray]
        List of coordinates.
    output_size : Tuple[int, int], optional
        Size of output arrays. Default is (256, 256).
    dilation_iterations : int, optional
        Number of iterations to dilate ground truth labels. Default is 1.
    coords_pad_length : Optional[int], optional
        Padded length of the coordinates sequence. Default is None.

    Returns
    -------
    deltas : jax.Array
        Array where each pixel is a vector to the nearest point in `coords`.
    labels : jax.Array
        Array where each pixel is a boolean for whether it contains a point in `coords`.

    References
    ----------
    .. [1] Laubscher, Emily, et al. "vanvalenlab/deepcell-spots: Deep Learning Library for Spot Detection." GitHub,
           https://github.com/vanvalenlab/deepcell-spots.
    """

    # Initialize the deltas and labels arrays.
    batch_size = len(coords)
    deltas = np.zeros((batch_size, *output_size, 2), dtype=float)
    labels = np.zeros((batch_size, *output_size), dtype=bool)

    # Determine the padded length of the coordinates sequence.
    coords_max_length = np.max([len(coord) for coord in coords])
    if (coords_pad_length is None) or (coords_pad_length < coords_max_length):
        coords_pad_length = coords_max_length

    # Generate ranges.
    i, j = np.arange(output_size[0]), np.arange(output_size[1])

    # Generate the dilation structuring element.
    structure = ndimage.generate_binary_structure(2, 2)

    for k, coord in enumerate(coords):

        # Remove coordinates outside the output arrays.
        coord = coord[(coord[:, 0] > -0.5) & (coord[:, 0] < output_size[0] - 0.5) &
                      (coord[:, 1] > -0.5) & (coord[:, 1] < output_size[1] - 0.5)]

        # Generate the labels array.
        rounded_coords = np.rint(coord).astype(int)
        labels[k][rounded_coords[:, 0], rounded_coords[:, 1]] = True

        # Apply the Euclidean distance transform on the labels array.
        edt_indices = ndimage.distance_transform_edt(~labels[k], return_distances=False, return_indices=True)

        # Pad coordinates to the same length.
        padding = ((0, coords_pad_length - len(coord)), (0, 0))
        coord = np.pad(coord, padding, constant_values=-1)

        # Apply the vectorized distance transform function.
        deltas[k] = vmap_vt(coord, edt_indices, i, j)

        # Dilate the labels array if necessary.
        if dilation_iterations > 0:
            labels[k] = ndimage.binary_dilation(labels[k], structure=structure, iterations=dilation_iterations)

    # Expand the shape of the labels array.
    labels = np.expand_dims(labels, axis=-1)

    return deltas, labels


@jit
def _vt(
        coord: jax.Array,
        edt_index: jax.Array,
        i: int,
        j: int
):

    """Voronoi transform for a single pixel.

    Parameters
    ----------
    coord : jax.Array
        List of coordinates.
    edt_index : jax.Array
        Euclidean distance transform index.
    i : int
        Pixel row index.
    j : int
        Pixel column index.

    Returns
    -------
    delta : jax.Array
        Vector to the nearest point.
    """

    distances = jnp.linalg.norm(coord - edt_index, axis=-1)
    delta = coord[jnp.argmin(distances)] - jnp.array((i, j))

    return delta


vmap_vt = vmap(vmap(_vt, in_axes=(None, 1, None, 0)), in_axes=(None, 1, 0, None))



from pathlib import Path

# Define Piscis paths.
PISCIS_DIR = Path.home() / '.piscis'
CACHE_DIR = PISCIS_DIR / 'cache'
CHECKPOINTS_DIR = PISCIS_DIR / 'checkpoints'
MODELS_DIR = PISCIS_DIR / 'models'

# Define Hugging Face paths.
HF_DATASETS_DIR = 'datasets/wniu/Piscis/'
HF_MODELS_DIR = 'wniu/Piscis/models/'



import deeptile
import jax
import numpy as np

from jax import random
from pathlib import Path
from typing import Dict, List, Optional, Tuple

from piscis.utils import pad, remove_duplicate_coords
from piscis.transforms import batch_adjust, RandomAugment, voronoi_transform


def generate_dataset(
        path: str,
        images: List[np.ndarray],
        coords: List[np.ndarray],
        key: jax.Array,
        tile_size: Tuple[int, int] = (256, 256),
        min_spots: int = 1,
        train_size: float = 0.70,
        test_size: float = 0.15
) -> None:

    """Generate a dataset from images and spot coordinates.

    Parameters
    ----------
    path : str
        Path to save dataset.
    images : List[np.ndarray]
        List of images.
    coords : List[np.ndarray]
        List of ground truth spot coordinates.
    key : jax.Array
        Random key used for splitting the dataset into training, validation, and test sets.
    tile_size : Tuple[int, int], optional
        Tile size used for splitting images. Default is (256, 256).
    min_spots : int, optional
        Minimum number of spots per tile. Default is 1.
    train_size : float, optional
        Fraction of dataset used for training. Default is 0.70.
    test_size : float, optional
        Fraction of dataset used for testing. Default is 0.15.
    """

    # Remove duplicate coordinates.
    for i in range(len(coords)):
        coords[i] = remove_duplicate_coords(coords[i])

    tiled_images_list = []
    tiled_coords_list = []

    for image, c in zip(images, coords):

        # Create a DeepTile object and get tiles.
        dt = deeptile.load(image)
        tiles1 = dt.get_tiles(tile_size, (0, 0))
        tiles2 = tiles1.import_data(c, 'coords')
        nonempty_tiles = [(tile1.compute(), tile2)
                          for tile1, tile2 in zip(tiles1[tiles1.nonempty_indices], tiles2[tiles2.nonempty_indices])
                          if len(tile2) >= min_spots]
        tiles1, tiles2 = zip(*nonempty_tiles)
        tiled_images_list += tiles1
        tiled_coords_list += tiles2

    tiled_images = np.empty(len(tiled_images_list), dtype=object)
    tiled_coords = np.empty(len(tiled_coords_list), dtype=object)
    tiled_images[:] = tiled_images_list
    tiled_coords[:] = tiled_coords_list

    # Randomly shuffle the tiles.
    size = len(tiled_images)
    perms = np.asarray(random.permutation(key, size))
    tiled_images = tiled_images[perms]
    tiled_coords = tiled_coords[perms]

    # Split the dataset into training, validation, and test sets.
    split_indices = np.rint(np.cumsum((train_size, test_size)) * size).astype(int)
    x_train = tiled_images[:split_indices[0]]
    y_train = tiled_coords[:split_indices[0]]
    x_valid = tiled_images[split_indices[1]:]
    y_valid = tiled_coords[split_indices[1]:]
    x_test = tiled_images[split_indices[0]:split_indices[1]]
    y_test = tiled_coords[split_indices[0]:split_indices[1]]

    # Create the dataset dictionary.
    np.savez(path, x_train=x_train, y_train=y_train, x_valid=x_valid, y_valid=y_valid, x_test=x_test, y_test=y_test)


def load_datasets(
        path: str,
        adjustment: Optional[str] = 'standardize',
        load_train: bool = True,
        load_valid: bool = True,
        load_test: bool = True
) -> Dict:

    """Load datasets from a directory or file.

    Parameters
    ----------
    path : str
        Path to a dataset or directory of datasets.
    adjustment : Optional[str], optional
        Adjustment type applied to images. Supported types are 'normalize' and 'standardize'. Default is 'standardize'.
    load_train : bool, optional
        Whether to load the training set. Default is True.
    load_valid : bool, optional
        Whether to load the validation set. Default is True.
    load_test : bool, optional
        Whether to load the test set. Default is True.

    Returns
    -------
    dataset : Dict
        Dataset dictionary.
    """

    # Create empty dictionaries.
    train = {'images': [], 'coords': []}
    valid = {'images': [], 'coords': []}
    test = {'images': [], 'coords': []}
    dataset = {}

    # Get dataset paths.
    path = Path(path)
    if path.is_file() and path.suffix == '.npz':
        dataset_paths = [path]
    else:
        dataset_paths = path.glob('*.npz')

    # Load datasets.
    for dataset_path in dataset_paths:
        npz = np.load(dataset_path, allow_pickle=True)
        if load_train:
            train['images'].append(npz['x_train'])
            train['coords'].append(npz['y_train'])
        if load_valid:
            valid['images'].append(npz['x_valid'])
            valid['coords'].append(npz['y_valid'])
        if load_test:
            test['images'].append(npz['x_test'])
            test['coords'].append(npz['y_test'])

    # Combine datasets and adjust images if necessary.
    if load_train:
        train['images'] = np.concatenate(train['images'])
        train['images'] = batch_adjust(train['images'], adjustment)
        train['coords'] = np.concatenate(train['coords'])
        dataset['train'] = train
    if load_valid:
        valid['images'] = np.concatenate(valid['images'])
        valid['images'] = batch_adjust(valid['images'], adjustment)
        valid['coords'] = np.concatenate(valid['coords'])
        dataset['valid'] = valid
    if load_test:
        test['images'] = np.concatenate(test['images'])
        test['images'] = batch_adjust(test['images'], adjustment)
        test['coords'] = np.concatenate(test['coords'])
        dataset['test'] = test

    return dataset


def transform_subdataset(
        subds: Dict,
        input_size: Tuple[int, int],
        min_spots: int = 1,
        key: Optional[jax.Array] = None
) -> Dict:

    """Transform subdataset for model training and validation.

    Parameters
    ----------
    subds : Dict
        Subdataset dictionary.
    input_size : Tuple[int, int]
        Size of the input images used during training.
    min_spots : int, optional
        Minimum number of spots per image. Default is 1.
    key : Optional[jax.Array], optional
        Random key used for data augmentation. Default is None.

    Returns
    -------
    transformed_subds : Dict
        Transformed subdataset dictionary.
    """

    if key is not None:

        base_scales = np.ones(len(subds['images']))

        # Create random augmentation transformer.
        transformer = RandomAugment()
        transformer.generate_transforms(subds['images'], key, base_scales, input_size)

        # Apply transformations.
        images = transformer.apply_image_transforms(subds['images'], interpolation='bilinear')
        coords_list = transformer.apply_coord_transforms(subds['coords'], filter_coords=True)

    else:

        images = pad(subds['images'])
        coords_list = list(subds['coords'])

    # Remove images with less than min_spots.
    counts = np.array([len(coord) for coord in coords_list])
    pop_indices = np.where((counts < min_spots))[0]
    for index in np.flip(pop_indices):
        images.pop(index)
        coords_list.pop(index)
    coords = np.empty(len(coords_list), dtype=object)
    coords[:] = coords_list

    # Created the transformed subdataset dictionary.
    transformed_subds = {
        'images': np.array(images),
        'coords': coords,
    }

    return transformed_subds


def transform_batch(
        batch: Dict,
        dilation_iterations: int = 1,
        coords_pad_length: Optional[int] = None
):

    """Transform batch for model training and validation.

    Parameters
    ----------
    batch : Dict
        Batch dictionary.
    dilation_iterations : int, optional
        Number of iterations to dilate ground truth labels. Default is 1.
    coords_pad_length : Optional[int], optional
        Padded length of the coordinates sequence. Default is None.
    """

    images = batch['images']
    coords = batch['coords']
    output_shape = images.shape[1:3]

    # Apply the Voronoi transform.
    deltas, labels = voronoi_transform(coords, output_shape, dilation_iterations, coords_pad_length)

    # Create the transformed batch dictionary.
    transformed_batch = {
        'images': images,
        'deltas': deltas,
        'labels': labels
    }

    return transformed_batch



import copy
import jax
import math

from flax import linen as nn
from functools import partial
from typing import Any, Callable, Dict, Optional, Sequence, Union

from piscis.networks.conv import MBConv, FusedMBConv
from piscis.networks.efficientnetv2_defaults import DEFAULT_BLOCKS_ARGS

ModuleDef = Any


def round_features(
        features: int,
        width_coefficient: float,
        min_depth: int,
        depth_divisor: int
) -> int:

    """Round the number of features based on the depth multiplier.

    Parameters
    ----------
    features : int
        Number of features.
    width_coefficient : float
        Scaling coefficient for network width.
    min_depth : int
        Minimum number of filters.
    depth_divisor : int
        Unit of network width.

    Returns
    -------
    new_features : int
        Rounded number of features.
    """

    features *= width_coefficient
    minimum_depth = min_depth or depth_divisor
    new_features = max(
        minimum_depth,
        int(features + depth_divisor / 2) // depth_divisor * depth_divisor,
    )
    new_features = int(new_features)

    return new_features


def round_repeats(
        repeats: int,
        depth_coefficient: float
) -> int:

    """Round number of repeats based on depth multiplier.

    Parameters
    ----------
    repeats : int
        Number of repeats.
    depth_coefficient : float
        Scaling coefficient for network depth.

    Returns
    -------
    new_repeats : int
        Rounded number of repeats.
    """

    new_repeats = int(math.ceil(depth_coefficient * repeats))

    return new_repeats


class EfficientNetV2(nn.Module):

    """EfficientNetV2 architecture.

    Attributes
    ----------
    stem_strides : int
        Strides of the stem convolution.
    blocks_args : Sequence
        List of arguments to construct block modules.
    dropout_rate : float
        Dropout rate at skip connections.
    bn_momentum : float
        Momentum parameter for batch norm layers.
    conv : ModuleDef
        Convolution module.
    dropout : ModuleDef
        Dropout module.
    bn : ModuleDef
        Batch norm module.
    act : Callable
        Activation function.
    """

    stem_strides: int
    blocks_args: Sequence
    dropout_rate: float
    bn_momentum: float
    conv: ModuleDef = nn.Conv
    dropout: ModuleDef = nn.Dropout
    bn: ModuleDef = nn.BatchNorm
    act: Callable = nn.swish

    @nn.compact
    def __call__(
            self,
            x: jax.Array,
            train: bool = True,
            capture_list: Optional[Sequence[int]] = None
    ) -> Union[jax.Array, Dict[int, jax.Array]]:

        # Initialize capture list.
        captures = {}

        bn = partial(
            self.bn,
            use_running_average=not train,
            momentum=self.bn_momentum
        )

        # Build stem.
        x = self.conv(
            features=self.blocks_args[0][0]['features_in'],
            kernel_size=(3, 3),
            strides=self.stem_strides,
            padding='SAME',
            use_bias=False,
            name='stem_conv',
        )(x)
        x = bn(
            name='stem_bn'
        )(x)
        x = self.act(x)

        # Build blocks.
        blocks_args = copy.deepcopy(self.blocks_args)
        b = 0
        blocks = float(sum(len(block_args) for block_args in blocks_args))

        for i, block_args in enumerate(blocks_args):

            for args in block_args:

                args = args.unfreeze()
                block = {0: MBConv, 1: FusedMBConv}[args.pop('conv_type')]

                if self.dropout_rate > 0:
                    dropout = partial(
                        self.dropout,
                        rate=self.dropout_rate * b / blocks,
                        deterministic=not train
                    )
                else:
                    dropout = None

                pool = args.pop('pool')
                if pool == 'avg':
                    x = nn.avg_pool(x, window_shape=(2, 2), strides=(2, 2))
                elif pool == 'max':
                    x = nn.max_pool(x, window_shape=(2, 2), strides=(2, 2))

                x = block(
                    conv=self.conv,
                    dropout=dropout,
                    bn=bn,
                    act=self.act,
                    **args,
                )(x)
                b += 1

            # Capture intermediate output if necessary.
            if (capture_list is not None) and (i in capture_list):
                captures[i] = x

        if capture_list is None:
            output = x
        else:
            output = captures

        return output


def build_efficientnetv2(
        blocks_args: Sequence[Dict[str, Any]],
        model_name: str,
        width_coefficient: float,
        depth_coefficient: float,
        stem_strides: int = 2,
        dropout_rate: float = 0.2,
        bn_momentum: float = 0.9,
        depth_divisor: int = 8,
        min_depth: int = 8,
        conv: ModuleDef = nn.Conv,
        dropout: ModuleDef = nn.Dropout,
        bn: ModuleDef = nn.BatchNorm,
        act: Callable = nn.swish
) -> partial:

    """Build EfficientNetV2 architecture.

    Parameters
    ----------
    blocks_args : Sequence[Dict[str, Any]]
        List of dictionaries of arguments to construct block modules.
    model_name : str
        Model name.
    width_coefficient : float
        Scaling coefficient for network width.
    depth_coefficient : float
        Scaling coefficient for network depth.
    stem_strides : int, optional
        Strides of the stem convolution. Default is 2.
    dropout_rate : float, optional
        Dropout rate at skip connections. Default is 0.2.
    bn_momentum : float, optional
        Momentum parameter for batch norm layers. Default is 0.9.
    depth_divisor : int, optional
        Unit of network width. Default is 8.
    min_depth : int, optional
        Minimum number of filters. Default is 8.
    conv : ModuleDef, optional
        Convolution module. Default is nn.Conv.
    dropout : ModuleDef, optional
        Dropout module. Default is nn.Dropout.
    bn : ModuleDef, optional
        Batch norm module. Default is nn.BatchNorm.
    act : Callable, optional
        Activation function. Default is nn.swish.

    Returns
    -------
    model : partial
        EfficientNetV2 architecture.
    """

    blocks_args = copy.deepcopy(blocks_args)

    long_blocks_args = []
    b = 0
    for i, block_args in enumerate(blocks_args):

        long_block_args = []

        assert block_args['num_repeat'] > 0

        # Update block input and output features based on depth multiplier.
        block_args['features_in'] = round_features(
            features=block_args['features_in'],
            width_coefficient=width_coefficient,
            min_depth=min_depth,
            depth_divisor=depth_divisor
        )
        block_args['features_out'] = round_features(
            features=block_args['features_out'],
            width_coefficient=width_coefficient,
            min_depth=min_depth,
            depth_divisor=depth_divisor
        )

        repeats = round_repeats(
            repeats=block_args.pop('num_repeat'),
            depth_coefficient=depth_coefficient
        )
        for j in range(repeats):

            # The first block needs to take care of stride and filter size increases.
            if j > 0:
                block_args['pool'] = None
                block_args['strides'] = 1
                block_args['features_in'] = block_args['features_out']

            long_block_args.append(block_args.copy())

            b += 1

        long_blocks_args.append(long_block_args)

    model = partial(
        EfficientNetV2,
        stem_strides=stem_strides,
        blocks_args=long_blocks_args,
        dropout_rate=dropout_rate,
        bn_momentum=bn_momentum,
        conv=conv,
        dropout=dropout,
        bn=bn,
        act=act,
        name=model_name
    )

    return model


EfficientNetV2B0 = build_efficientnetv2(
    blocks_args=DEFAULT_BLOCKS_ARGS['efficientnetv2-b0'],
    model_name='efficientnetv2-b0',
    width_coefficient=1.0,
    depth_coefficient=1.0,
)

EfficientNetV2B1 = build_efficientnetv2(
    blocks_args=DEFAULT_BLOCKS_ARGS['efficientnetv2-b1'],
    model_name='efficientnetv2-b1',
    width_coefficient=1.0,
    depth_coefficient=1.1,
)

EfficientNetV2B2 = build_efficientnetv2(
    blocks_args=DEFAULT_BLOCKS_ARGS['efficientnetv2-b2'],
    model_name='efficientnetv2-b2',
    width_coefficient=1.1,
    depth_coefficient=1.2,
)

EfficientNetV2B3 = build_efficientnetv2(
    blocks_args=DEFAULT_BLOCKS_ARGS['efficientnetv2-b3'],
    model_name='efficientnetv2-b3',
    width_coefficient=1.2,
    depth_coefficient=1.4,
)

EfficientNetV2S = build_efficientnetv2(
    blocks_args=DEFAULT_BLOCKS_ARGS['efficientnetv2-s'],
    model_name='efficientnetv2-s',
    width_coefficient=1.0,
    depth_coefficient=1.0,
)

EfficientNetV2M = build_efficientnetv2(
    blocks_args=DEFAULT_BLOCKS_ARGS['efficientnetv2-m'],
    model_name='efficientnetv2-m',
    width_coefficient=1.0,
    depth_coefficient=1.0,
)

EfficientNetV2L = build_efficientnetv2(
    blocks_args=DEFAULT_BLOCKS_ARGS['efficientnetv2-l'],
    model_name='efficientnetv2-l',
    width_coefficient=1.0,
    depth_coefficient=1.0,
)
